import ubelt as ub
from xcookie.builders import common_ci


def build_gitlab_ci(self):
    """
    Example:
        >>> from xcookie.builders.gitlab_ci import *  # NOQA
        >>> from xcookie.main import XCookieConfig
        >>> from xcookie.main import TemplateApplier
        >>> config = XCookieConfig(tags=['purepy'])
        >>> self = TemplateApplier(config)
        >>> text = build_gitlab_ci(self)
        >>> print(text)
    """
    if 'purepy' in self.tags:
        return make_purepy_ci_jobs(self)
    else:
        raise NotImplementedError


def build_gitlab_rules(self):
    group = self.remote_info['group']
    repo_name = self.remote_info['repo_name']
    text = ub.codeblock(
        f'''
        # Rules for where jobs can run
        # Derived from: https://gitlab.kitware.com/cmake/cmake/-/blob/v3.25.1/.gitlab/rules.yml
        # For an overview of gitlab rules see:
        # https://docs.gitlab.com/ee/ci/yaml/#workflowrules

        .run_manually:
            rules:
                - if: '$CI_MERGE_REQUEST_ID'
                  when: manual
                - if: '$CI_COMMIT_REF_PROTECTED == true'
                  when: on_success
                - if: '$CI_PROJECT_PATH == "{group}/{repo_name}" && $CI_PIPELINE_SOURCE == "schedule"'
                  when: on_success
                - if: '$CI_PROJECT_PATH == "{group}/{repo_name}"'
                  when: manual
                - when: never

        .run_automatically:
            rules:
                - if: '$CI_MERGE_REQUEST_ID'
                  when: on_success
                - if: '$CI_PROJECT_PATH == "{group}/{repo_name}" && $CI_PIPELINE_SOURCE == "schedule"'
                  when: on_success
                - if: '$CI_PROJECT_PATH == "{group}/{repo_name}"'
                  when: delayed
                  start_in: 5 minutes
                - when: never

        .run_dependent:
            rules:
                - if: '$CI_MERGE_REQUEST_ID'
                  when: on_success
                - if: '$CI_PROJECT_PATH == "{group}/{repo_name}"'
                  when: on_success
                - when: never
        ''')
    return text


# class YamlBuilder:
#     def __init__(self, data):
#         pass


def make_purepy_ci_jobs(self):
    import ruamel.yaml
    from ruamel.yaml.comments import CommentedMap, CommentedSeq
    from xcookie.util_yaml import Yaml

    enable_gpg = self.config['enable_gpg']

    body = CommentedMap()
    body['stages'] = CommentedSeq([
        'build',
        'test',
    ])
    body.yaml_add_eol_comment('stages', 'TEMPLATE1,c 1')

    common_template = ub.udict(Yaml.loads(ub.codeblock(
        '''
        tags:
            # Tags define which runners will accept which jobs
            - docker
            - linux-x86_64
            - build

        variables:
            # Change pip's cache directory to be inside the project directory
            # since we can only cache local items.
            PIP_CACHE_DIR: "$CI_PROJECT_DIR/.cache/pip"

        except:
            # Don't run the pipeline for new tags
            - tags

        cache:
            paths:
                - .cache/pip
        ''')))

    common_template = CommentedMap(common_template)
    common_template.yaml_set_anchor('common_template')
    body['.common_template'] = common_template

    if 'kitware' in self.tags:
        body['.common_template']['tags'].append('kitware-python-stack')

    wheelhouse_dpath = 'dist'
    build_wheel_parts = common_ci.make_build_wheel_parts(self, wheelhouse_dpath)
    build_template = {
        'stage': 'build',
        'before_script': [
            'python -V  # Print out python version for debugging',
        ],
        'script': build_wheel_parts['commands'],
        'artifacts': {
            'paths': [
                build_wheel_parts['artifact'],
            ]
        },
    }
    build_template = CommentedMap(build_template)
    build_template.yaml_set_anchor('build_template')
    body['.build_template'] = build_template
    build_template.add_yaml_merge([(0, common_template)])

    common_test_template = {
        'stage': 'test',

        # Coverage is a regex that will parse the coverage from the test stdout
        'coverage': '/TOTAL.+ ([0-9]{1,3}%)/',
    }

    common_test_template = CommentedMap(common_test_template)
    common_test_template.yaml_set_anchor('common_test_template')
    common_test_template.add_yaml_merge([(0, common_template)])
    body['.common_test_template'] = common_test_template

    setup_venv_template = Yaml.CodeBlock(
        '''
        # Setup the correct version of python (which should be the same as this instance)
        python --version  # Print out python version for debugging
        export PYVER=$(python -c "import sys; print('{}{}'.format(*sys.version_info[0:2]))")
        python -m pip install virtualenv
        python -m virtualenv venv$PYVER
        source venv$PYVER/bin/activate
        pip install pip -U
        pip install pip setuptools -U
        pip install pygments
        python --version  # Print out python version for debugging
        ''')

    test_templates = {}
    loose_cv2 = ''
    strict_cv2 = ''
    if 'cv2' in self.tags:
        loose_cv2 = ',headless'
        strict_cv2 = ',headless-strict'
    install_extras = ub.udict({
        'minimal-loose'  : 'tests' + loose_cv2,
        'full-loose'     : 'tests,optional' + loose_cv2,
        'minimal-strict' : 'tests-strict,runtime-strict' + strict_cv2,
        'full-strict'    : 'tests-strict,runtime-strict,optional-strict' + strict_cv2,
    })
    for extra_key, extra in install_extras.items():
        if 'gdal' in self.tags:
            special_install_lines = [
                # TODO: handle strict
                'pip install -r requirements/gdal.txt',
            ]
        else:
            special_install_lines = []
        workspace_dname = 'sandbox'
        install_and_test_wheel_parts = common_ci.make_install_and_test_wheel_parts(
            self, wheelhouse_dpath, special_install_lines, workspace_dname)
        test_steps = [
            f'export INSTALL_EXTRAS="{extra}"',
        ]
        test_steps += install_and_test_wheel_parts['install_wheel_commands']
        test_steps += install_and_test_wheel_parts['test_wheel_commands']
        test = {
            'before_script': [setup_venv_template],
            'script': test_steps,
        }
        anchor = f'test_{extra_key}_template'
        test = CommentedMap(test)
        test.yaml_set_anchor(anchor)
        test.add_yaml_merge([(0, common_test_template)])
        body['.' + anchor] = test
        test_templates[extra_key] = test

    python_images = {
        'cp311': 'python:3.11',
        'cp310': 'python:3.10',
        'cp39': 'python:3.9',
        'cp38': 'python:3.8',
        'cp37': 'python:3.7',
        'cp36': 'python:3.6',
    }

    build_names = []

    jobs = {}
    opsys = 'linux'
    arch = 'x86_64'
    for pyver in self.config['ci_cpython_versions']:
        cpver = 'cp' + pyver.replace('.', '')
        if cpver in python_images:
            swenv_key = f'{cpver}-{opsys}-{arch}'  # software environment key
            build_name = f'build/{swenv_key}'
            build_names.append(build_name)

            build_job = {
                'image': python_images[cpver],
            }
            build_job = CommentedMap(build_job)
            build_job.add_yaml_merge([(0, build_template)])
            jobs[build_name] = build_job

            for extra_key, common_test_template in test_templates.items():
                test_name = f'test/{extra_key}/{swenv_key}'
                test_job = {
                    'image': python_images[cpver],
                    'needs': [
                        build_name,
                    ]
                }
                test_job = CommentedMap(test_job)
                test_job.add_yaml_merge([(0, common_test_template)])
                jobs[test_name] = test_job

    body.update(jobs)

    deploy_image = python_images['cp38']

    if enable_gpg:
        gpgsign_job = {}
        gpgsign_job.update(ub.udict(Yaml.loads(ub.codeblock(
            f'''
            image:
                {deploy_image}

            stage:
                gpgsign

            artifacts:
                paths:
                    - {wheelhouse_dpath}/*.asc

            only:
                refs:
                    # Gitlab will only expose protected variables on protected branches
                    # (which I've set to be main and release), so only run this stage
                    # there.
                    - master
                    - main
                    - release
            '''))))

        gpgsign_job['needs'] = [{'job': build_name, 'artifacts': True} for build_name in build_names]

        gpgsign_job.update(Yaml.loads(ub.codeblock(
            '''
            script:
                - ls ''' + wheelhouse_dpath + '''
                - export GPG_EXECUTABLE=gpg
                - export GPG_KEYID=$(cat dev/public_gpg_key)
                - echo "GPG_KEYID = $GPG_KEYID"
                # Decrypt and import GPG Keys / trust
                # note the variable pointed to by VARNAME_CI_SECRET is a protected variables only available on main and release branch
                - source dev/secrets_configuration.sh
                - CI_SECRET=${!VARNAME_CI_SECRET}
                - $GPG_EXECUTABLE --version
                - openssl version
                - $GPG_EXECUTABLE --list-keys
                # note CI_KITWARE_SECRET is a protected variables only available on main and release branch
                - CIS=$CI_SECRET openssl enc -aes-256-cbc -pbkdf2 -md SHA512 -pass env:CIS -d -a -in dev/ci_public_gpg_key.pgp.enc | $GPG_EXECUTABLE --import
                - CIS=$CI_SECRET openssl enc -aes-256-cbc -pbkdf2 -md SHA512 -pass env:CIS -d -a -in dev/gpg_owner_trust.enc | $GPG_EXECUTABLE --import-ownertrust
                - CIS=$CI_SECRET openssl enc -aes-256-cbc -pbkdf2 -md SHA512 -pass env:CIS -d -a -in dev/ci_secret_gpg_subkeys.pgp.enc | $GPG_EXECUTABLE --import
                - GPG_SIGN_CMD="$GPG_EXECUTABLE --batch --yes --detach-sign --armor --local-user $GPG_KEYID"
            ''')))
        gpgsign_job['script'].append(
            Yaml.CodeBlock(
                '''
                WHEEL_PATHS=(''' + wheelhouse_dpath + '''/*.whl)
                WHEEL_PATHS_STR=$(printf '"%s" ' "${WHEEL_PATHS[@]}")
                echo "$WHEEL_PATHS_STR"
                for WHEEL_PATH in "${WHEEL_PATHS[@]}"
                do
                    echo "------"
                    echo "WHEEL_PATH = $WHEEL_PATH"
                    $GPG_SIGN_CMD --output $WHEEL_PATH.asc $WHEEL_PATH
                    $GPG_EXECUTABLE --verify $WHEEL_PATH.asc $WHEEL_PATH  || echo "hack, the first run of gpg very fails"
                    $GPG_EXECUTABLE --verify $WHEEL_PATH.asc $WHEEL_PATH
                done
                '''
            )
        )
        gpgsign_job['script'].append(f'ls {wheelhouse_dpath}')

        gpgsign_job = CommentedMap(gpgsign_job)
        gpgsign_job.add_yaml_merge([(0, common_template)])
        body['stages'].append('gpgsign')
        body['gpgsign/wheels'] = gpgsign_job

    deploy = True
    if deploy:
        deploy_job = {}
        deploy_job.update(ub.udict(Yaml.loads(ub.codeblock(
            f'''
            image:
                {deploy_image}

            stage:
                deploy

            only:
                refs:
                    - release
            '''))))
        deploy_script = [
            'pip install pyopenssl ndg-httpsclient pyasn1 requests[security] twine -U',
            f'ls {wheelhouse_dpath}',
        ]
        deploy_script += [
            Yaml.CodeBlock(
                '''
                WHEEL_PATHS=(''' + wheelhouse_dpath + '''/*.whl)
                WHEEL_PATHS_STR=$(printf '"%s" ' "${WHEEL_PATHS[@]}")
                source dev/secrets_configuration.sh
                TWINE_PASSWORD=${!VARNAME_TWINE_PASSWORD}
                TWINE_USERNAME=${!VARNAME_TWINE_USERNAME}
                echo "$WHEEL_PATHS_STR"
                for WHEEL_PATH in "${WHEEL_PATHS[@]}"
                do
                    twine check $WHEEL_PATH.asc $WHEEL_PATH
                    twine upload --username $TWINE_USERNAME --password $TWINE_PASSWORD $WHEEL_PATH.asc $WHEEL_PATH || echo "upload already exists"
                done
                ''')
        ]
        deploy_script += [
            Yaml.CodeBlock(
                r'''
                # Have the server git-tag the release and push the tags
                export VERSION=$(python -c "import setup; print(setup.VERSION)")
                # do sed twice to handle the case of https clone with and without a read token
                URL_HOST=$(git remote get-url origin | sed -e 's|https\?://.*@||g' | sed -e 's|https\?://||g' | sed -e 's|git@||g' | sed -e 's|:|/|g')
                source dev/secrets_configuration.sh
                CI_SECRET=${!VARNAME_CI_SECRET}
                PUSH_TOKEN=${!VARNAME_PUSH_TOKEN}
                echo "URL_HOST = $URL_HOST"
                # A git config user name and email is required. Set if needed.
                if [[ "$(git config user.email)" == "" ]]; then
                    git config user.email "ci@gitlab.org.com"
                    git config user.name "Gitlab-CI"
                fi
                TAG_NAME="v${VERSION}"
                echo "TAG_NAME = $TAG_NAME"
                if [ $(git tag -l "$TAG_NAME") ]; then
                    echo "Tag already exists"
                else
                    # if we messed up we can delete the tag
                    # git push origin :refs/tags/$TAG_NAME
                    # and then tag with -f
                    git tag $TAG_NAME -m "tarball tag $VERSION"
                    git push --tags "https://git-push-token:${PUSH_TOKEN}@${URL_HOST}"
                fi
                ''')
        ]
        deploy_job['script'] = deploy_script
        deploy_job = CommentedMap(deploy_job)
        deploy_job.add_yaml_merge([(0, common_template)])
        body['stages'].append('deploy')
        body['deploy/wheels'] = deploy_job

    body_text = ruamel.yaml.round_trip_dump(body, Dumper=ruamel.yaml.RoundTripDumper)

    body = ruamel.yaml.round_trip_load(body_text)
    body['stages'].yaml_set_start_comment('TEMPLATES')
    # body.yaml_set_comment_before_after_key('stages', 'before test1 (top level)', after='before test2\n\n')
    # body.yaml_add_eol_comment('STAGE COMMENT', key='stages')
    body_text = ruamel.yaml.round_trip_dump(body, Dumper=ruamel.yaml.RoundTripDumper)

    header = ub.codeblock(
        '''
        # Autogenerated by ~/code/xcookie/xcookie/builders/gitlab_ci.py
        '''
    )
    footer = '# end'
    text = header + '\n\n' + body_text + '\n\n' + footer
    return text



import ubelt as ub


class BadgeBuilder:
    def __init__(self):
        self.badges = {}

    def build_templates(self, group, repo_name, main_branch, workflow):
        repo_dashname = repo_name.replace('_', '-')
        badges = self.badges = {}
        badges['CircleCI'] = {
            'image': f'https://circleci.com/gh/{group}/{repo_name}.svg?style=svg',
            'target': f'https://circleci.com/gh/{group}/{repo_name}',
        }
        badges['Appveyor'] = {
            'image': f'https://ci.appveyor.com/api/projects/status/github/{group}/{repo_name}?branch={main_branch}&svg=True',
            'target': f'https://ci.appveyor.com/project/{group}/{repo_name}/branch/{main_branch}',
        }
        badges['Codecov'] = {
            'image': f'https://codecov.io/github/{group}/{repo_name}/badge.svg?branch={main_branch}&service=github',
            'target': f'https://codecov.io/github/{group}/{repo_name}?branch={main_branch}',
        }
        badges['Pypi'] = {
            'image': f'https://img.shields.io/pypi/v/{repo_name}.svg',
            'target': f'https://pypi.python.org/pypi/{repo_name}',
        }
        badges['PypiDownloads'] = {
            'image': f'https://img.shields.io/pypi/dm/{repo_name}.svg',
            'target': f'https://pypistats.org/packages/{repo_name}',
        }
        badges['ReadTheDocs'] = {
            'image': f'https://readthedocs.org/projects/{repo_dashname}/badge/?version=latest',
            'target': f'http://{repo_dashname}.readthedocs.io/en/latest/',
        }
        badges['GithubActions'] = {
            'image': f'https://github.com/{group}/{repo_name}/actions/workflows/{workflow}.yml/badge.svg?branch={main_branch}',
            'target': f'https://github.com/{group}/{repo_name}/actions?query=branch%3A{main_branch}',
        }
        return badges


def build_readme(self):
    badges = {}
    remote_info = self.remote_info
    main_branch = 'main'
    group = remote_info['group']
    repo_name = remote_info['repo_name']

    if 'binpy' in self.config['tags']:
        workflow = 'test_binaries'
    else:
        workflow = 'tests'
    badges = BadgeBuilder().build_templates(group, repo_name, main_branch, workflow)

    # badges['CodeQuality'] = {
    #     'image': f'image:: https://api.codacy.com/project/badge/Grade/4d815305fc014202ba7dea09c4676343',
    #     'target': f'https://www.codacy.com/manual/{group}/{repo_name}?utm_source=github.com&amp;utm_medium=referral&amp;utm_content={group}/{repo_name}&amp;utm_campaign=Badge_Grade',
    # }

    parts = []

    title = f'The {repo_name} Module'
    title = title + '\n' + ('=' * len(title))
    parts.append(title)
    parts.append('')

    chosen_badges = ['Pypi', 'PypiDownloads']
    if 'github' in self.config['tags']:
        chosen_badges += ['GithubActions', 'Codecov']

    badge_header = ' '.join(['|{}|'.format(b) for b in chosen_badges])
    parts.append(badge_header)
    parts.append('')

    for b in chosen_badges:
        badge = badges[b]
        badge_def = ub.codeblock(
            f'''
            .. |{b}| image:: {badge['image']}
                :target: {badge['target']}
            '''
        )
        parts.append(badge_def)

    readme_text = '\n\n'.join(parts)
    return readme_text


def _ibeis_badges(repo_names):
    """
    repo_names = [
        'ibeis',
        'utool',
        'vtool_ibeis',
        'plottool_ibeis',
        'guitool_ibeis',
        'pyhesaff',
        'pyflann_ibeis',
        'vtool_ibeis_ext',
        'graphid',
    ]
    """
    group = 'Erotemic'
    main_branch = 'main'
    workflow = 'tests'
    lines = []
    for repo_name in repo_names:
        badges = BadgeBuilder().build_templates(group, repo_name, main_branch, workflow)
        b = 'GithubActions'
        kw = badges[b]
        badge_name = repo_name + b
        text = ub.codeblock(
            '''
            .. |{badge_name}| image:: {image}
                :target: {target}
            '''
        ).format(**kw, badge_name=badge_name)
        lines.append(text)

        print(f'|{badge_name}|')
    print('\n'.join(lines))



"""
Port and extension of xdoctest directives.
"""
import sys
import os
import re
import warnings
import operator
from xdoctest import utils
from xdoctest import static_analysis as static
from collections import namedtuple
# from xdoctest import exceptions


def named(key, pattern):
    """ helper for regex """
    return '(?P<{}>{})'.format(key, pattern)


Effect = namedtuple('Effect', ('action', 'key', 'value'))


def extract_directive_comment(source):
    """
    Different than the xdoctest version. Finds the last comment part of a line

    source = '# acommend  # b comment'
    list(extract_directive_comment(source))

    # TODO: lark grammar?
    """
    for comment in static.extract_comments(source):
        current = comment
        stripped = current[1:]
        while True:
            n = list(static.extract_comments(stripped))
            if len(n) == 0:
                yield current
                break
            else:
                current = n[0]
                stripped = current[1:]


class Directive(utils.NiceRepr):
    """
    Directives modify the runtime state.
    """
    def __init__(self, name, positive=True, args=[], inline=None):
        self.name = name
        self.args = args
        self.inline = inline
        self.positive = positive

    @classmethod
    def extract(cls, text, directive_re, commands):
        """
        Parses directives from a line or repl line

        Args:
            text (str):
                must correspond to exactly one PS1 line and its PS2 followups.

            prefix (str | None):
                The directive "namespace". If None, this uses the xdoctest
                defaults of ``DIRECTIVE_RE``. This will always be the case for
                xdoctest, but we are extending this class for use elsewhere.

        Yields:
            Directive: directive: the parsed directives
        """
        # Flag extracted directives as inline iff the text is only comments
        inline = not all(line.strip().startswith('#')
                         for line in text.splitlines())
        #
        for comment in extract_directive_comment(text):
            # remove the first comment character and see if the comment matches
            # the directive pattern
            m = directive_re.match(comment[1:].strip())
            if m:
                for key, optstr in m.groupdict().items():
                    if optstr:
                        optparts = _split_opstr(optstr)
                        # optparts = optstr.split(',')
                        for optpart in optparts:
                            directive = parse_directive_optstr(optpart, commands, inline)
                            if directive:
                                yield directive

    def __nice__(self):
        prefix = ['-', '+'][int(self.positive)]
        if self.args:
            argstr = ', '.join(self.args)
            return '{}{}({})'.format(prefix, self.name, argstr)
        else:
            return '{}{}'.format(prefix, self.name)

    def _unpack_args(self, num):
        from xdoctest.utils import util_deprecation
        util_deprecation.schedule_deprecation3(
            modname='xdoctest',
            name='Directive._unpack_args', type='method',
            migration='there is no need to use this',
            deprecate='1.0.0', error='1.1.0', remove='1.2.0'
        )
        nargs = self.args
        if len(nargs) != 1:
            raise TypeError(
                '{} directive expected exactly {} argument(s), '
                'got {}'.format(self.name, num, nargs))
        return self.args

    def effect(self, argv=None, environ=None):
        from xdoctest.utils import util_deprecation
        util_deprecation.schedule_deprecation3(
            modname='xdoctest',
            name='Directive.effect', type='method',
            migration='Use Directive.effects instead',
            deprecate='1.0.0', error='1.1.0', remove='1.2.0'
        )
        effects = self.effects(argv=argv, environ=environ)
        if len(effects) > 1:
            raise Exception('Old method cannot handle multiple effects')
        return effects[0]

    def effects(self, argv=None, environ=None):
        """
        Returns how this directive modifies a RuntimeState object

        This is called by :func:`RuntimeState.update` to update itself

        Args:
            argv (List[str], default=None):
                if specified, overwrite sys.argv
            environ (Dict[str, str], default=None):
                if specified, overwrite os.environ

        Returns:
            List[Effect]: list of named tuples containing:
                action (str): code indicating how to update
                key (str): name of runtime state item to modify
                value (object): value to modify with
        """
        key = self.name
        value = None

        effects = []
        if self.name == 'REQUIRES':
            # Special handling of REQUIRES
            for arg in self.args:
                value = arg
                if _is_requires_satisfied(arg, argv=argv, environ=environ):
                    # If the requirement is met, then do nothing,
                    action = 'noop'
                else:
                    # otherwise, add or remove the condition from REQUIREMENTS,
                    # depending on if the directive is positive or negative.
                    if self.positive:
                        action = 'set.add'
                    else:
                        action = 'set.remove'
                effects.append(Effect(action, key, value))
        elif key.startswith('REPORT_'):
            # Special handling of report style
            if self.positive:
                action = 'noop'
            else:
                action = 'set_report_style'
            effects.append(Effect(action, key, value))
        else:
            # The action overwrites state[key] using value
            action = 'assign'
            value = self.positive
            effects.append(Effect(action, key, value))
        return effects


def _split_opstr(optstr):
    """
    Simplified balanced paren logic to only split commas outside of parens

    Example:
        >>> optstr = '+FOO, REQUIRES(foo,bar), +ELLIPSIS'
        >>> _split_opstr(optstr)
        ['+FOO', 'REQUIRES(foo,bar)', '+ELLIPSIS']
    """
    import re
    stack = []
    split_pos = []
    for match in re.finditer(r',|\(|\)', optstr):
        token = match.group()
        if token == ',' and not stack:
            # Only split when there are no parens
            split_pos.append(match.start())
        elif token == '(':
            stack.append(token)
        elif token == ')':
            stack.pop()
    assert len(stack) == 0, 'parens not balanced'

    parts = []
    prev = 0
    for curr in split_pos:
        parts.append(optstr[prev:curr].strip())
        prev = curr + 1
    curr = None
    parts.append(optstr[prev:curr].strip())
    return parts


def _is_requires_satisfied(arg, argv=None, environ=None):
    """
    Determines if the argument to a REQUIRES directive is satisfied

    Args:
        arg (str): condition code
        argv (List[str]): cmdline if arg is cmd code usually ``sys.argv``
        environ (Dict[str, str]): environment variables usually ``os.environ``

    Returns:
        bool: flag - True if the requirement is met
    """
    # TODO: add python version options
    SYS_PLATFORM_TAGS = ['win32', 'linux', 'darwin', 'cywgin']
    OS_NAME_TAGS = ['posix', 'nt', 'java']
    PY_IMPL_TAGS = ['cpython', 'ironpython', 'jython', 'pypy']
    # TODO: tox tags: https://tox.readthedocs.io/en/latest/example/basic.html
    PY_VER_TAGS = ['py2', 'py3']

    arg_lower = arg.lower()

    if arg.startswith('-'):
        if argv is None:
            argv = sys.argv
        flag = arg in argv
    elif arg.startswith('module:'):
        parts = arg.split(':')
        if len(parts) != 2:
            raise ValueError('xdoctest module REQUIRES directive has too many parts')
        # set flag to False (aka SKIP) if the module does not exist
        modname = parts[1]
        flag = _module_exists(modname)
    elif arg.startswith('env:'):
        if environ is None:
            environ = os.environ
        parts = arg.split(':')
        if len(parts) != 2:
            raise ValueError('xdoctest env REQUIRES directive has too many parts')
        envexpr = parts[1]
        expr_parts = re.split('(==|!=|>=)', envexpr)
        if len(expr_parts) == 1:
            # Test if the environment variable is truthy
            env_key = expr_parts[0]
            flag = bool(environ.get(env_key, None))
        elif len(expr_parts) == 3:
            # Test if the environment variable is equal to an expression
            env_key, op_code, value = expr_parts
            env_val = environ.get(env_key, None)
            if op_code == '==':
                op = operator.eq
            elif op_code == '!=':
                op = operator.ne
            else:
                raise KeyError(op_code)
            flag = op(env_val, value)
        else:
            raise ValueError('Too many expr_parts={}'.format(expr_parts))
    elif arg_lower in SYS_PLATFORM_TAGS:
        flag = sys.platform.lower().startswith(arg_lower)
    elif arg_lower in OS_NAME_TAGS:
        flag = os.name.startswith(arg_lower)
    elif arg_lower in PY_IMPL_TAGS:
        import platform
        flag = platform.python_implementation().lower().startswith(arg_lower)
    elif arg_lower in PY_VER_TAGS:
        if sys.version_info[0] == 2:  # nocover
            flag = arg_lower == 'py2'
        elif sys.version_info[0] == 3:  # pragma: nobranch
            flag = arg_lower == 'py3'
        else:  # nocover
            flag = False
    else:
        msg = utils.codeblock(
            '''
            Argument to REQUIRES directive must be either
            (1) a PLATFORM or OS tag (e.g. win32, darwin, linux),
            (2) a command line flag prefixed with '--', or
            (3) a module prefixed with 'module:'.
            (4) an environment variable prefixed with 'env:'.
            Got arg={!r}
            ''').replace('\n', ' ').strip().format(arg)
        raise ValueError(msg)
    return flag


class DirectiveExtractor:
    """
    Example:
        >>> from xcookie.directive import *  # NOQA
        >>> namespace = 'xcookie'
        >>> commands = ['UNCOMMENT_IF', 'COMMENT_IF']
        >>> self = DirectiveExtractor(namespace, commands)
        >>> text = '- this line is not python # xcookie: +COMMENT_IF(cv2)'
        >>> text = '# commented line # xcookie: +UNCOMMENT_IF(cv2)'
        >>> extracted = self.extract(text)
        >>> assert len(extracted) == 1
        >>> directive = extracted[0]
    """
    def __init__(self, namespace, commands):
        self.commands = commands
        self.namesapce = namespace
        directive_patterns = [
            namespace + r':\s*' + named('style2', '.*'),
        ]
        directive_re = re.compile('|'.join(directive_patterns), flags=re.IGNORECASE)
        self.directive_re = directive_re

    def extract(self, text):
        extracted = list(Directive.extract(text, self.directive_re, self.commands))
        return extracted


_MODNAME_EXISTS_CACHE = {}


def _module_exists(modname):
    if modname not in _MODNAME_EXISTS_CACHE:
        from xdoctest import static_analysis as static
        modpath = static.modname_to_modpath(modname)
        exists_flag = modpath is not None
        _MODNAME_EXISTS_CACHE[modname] = exists_flag
    exists_flag = _MODNAME_EXISTS_CACHE[modname]
    return exists_flag


def parse_directive_optstr(optpart, commands, inline=None):
    """
    Parses the information in the directive from the "optpart"

    optstrs are:
        optionally prefixed with ``+`` (default) or ``-``
        comma separated
        may contain one paren enclosed argument (experimental)
        all spaces are ignored

    Returns:
        Directive: the parsed directive
    """
    optpart = optpart.strip()
    # all spaces are ignored
    optpart = optpart.replace(' ', '')

    paren_pos = optpart.find('(')
    if paren_pos > -1:
        # handle simple paren case.
        body = optpart[paren_pos + 1:optpart.find(')')]
        args = [a.strip() for a in body.split(',')]
        # args = [optpart[paren_pos + 1:optpart.find(')')]]
        optpart = optpart[:paren_pos]
    else:
        args = []

    # Determine if the option starts with + or - (we assume + by default)
    if optpart.startswith(('+', '-')):
        positive = not optpart.startswith('-')
        name = optpart[1:]
    else:
        positive = True
        name = optpart

    name = name.upper()
    if name not in commands:
        msg = 'Unknown directive: {!r}'.format(optpart)
        warnings.warn(msg)
    else:
        directive = Directive(name, positive, args, inline)
        return directive




def build_docs_index(self):
    import ubelt as ub
    parts = []
    tags = set(self.config['tags'])
    mod_name = self.config['mod_name']

    if {'gitlab', 'kitware'}.issubset(tags):
        parts.append(ub.codeblock(
            f'''
            :gitlab_url: https://gitlab.kitware.com/computer-vision/{mod_name}
            '''))

    logo_part = ub.codeblock(
        '''
        .. The large version wont work because github strips rst image rescaling. https://i.imgur.com/AcWVroL.png
            # TODO: Add a logo
            .. image:: https://i.imgur.com/PoYIsWE.png
               :height: 100px
               :align: left
        ''')

    parts.append(logo_part)

    title = f"Welcome to {mod_name}'s documentation!"
    underline = '=' * len(title)
    title_part = title + '\n' + underline
    parts.append(title_part)

    init_part = ub.codeblock(
        f'''
        .. The __init__ files contains the top-level documentation overview
        .. automodule:: {mod_name}.__init__
           :show-inheritance:
        ''')
    parts.append(init_part)

    if 0:
        usefulness_part = ub.codeblock(
            '''
            .. # Computed function usefulness
            .. include:: function_usefulness.rst
            ''')
        parts.append(usefulness_part)

    sidebar_part = ub.codeblock(
        f'''
        .. toctree::
           :maxdepth: 5

           {mod_name}


        Indices and tables
        ==================

        * :ref:`genindex`
        * :ref:`modindex`
        ''')
    parts.append(sidebar_part)

    text = '\n\n'.join(parts)
    return text



"""
The xcookie module is a CLI tool for initializing and maintaining standardized
repo infastructure. In other words it is a Python project cookie cutter that
attempts to help keep existing modules up to date with latest infastructure
developments.

Currently this module is specialized towards Erotemic / PyUtils / Kitware
projects but the goal is to eventually generalize everything.

+------------------+----------------------------------------------+
| Read the docs    | https://xcookie.readthedocs.io               |
+------------------+----------------------------------------------+
| Github           | https://github.com/Erotemic/xcookie          |
+------------------+----------------------------------------------+
| Pypi             | https://pypi.org/project/xcookie             |
+------------------+----------------------------------------------+

"""
__version__ = '0.2.1'



from rich import prompt
import ubelt as ub


class FuzzyPrompt(prompt.Prompt):
    """
    The user just needs to enter a non-ambiguous prefix
    """
    def process_response(self, value: str) -> str:
        """Normalize the response"""
        assert self.choices is not None
        got = value.strip().lower()
        norm_choices = [c.lower() for c in self.choices]
        flags = [c.startswith(got) for c in norm_choices]
        if sum(flags) != 1:
            raise prompt.InvalidResponse(self.validate_error_message)
        norm = ub.peek(ub.compress(self.choices, flags))
        return norm



"""
This is a Python script to apply the xcookie template to either create a new repo
or update an existing one with the latest standards.

TODO:
    Port logic from ~/misc/make_new_python_package_repo.sh

ComamndLine:
    ~/code/xcookie/xcookie/main.py

    python -m xcookie.main

ExampleUsage:
    # Update my repos
    python -m xcookie.main --repodir=$HOME/code/pyflann_ibeis --tags="erotemic,github,binpy"

    python -m xcookie.main --repodir=$HOME/code/whodat --tags="kitware,gitlab,purepy,cv2,gdal"
    python -m xcookie.main --repodir=$HOME/code/whatdat --tags="kitware,gitlab,purepy,cv2,gdal"
    python -m xcookie.main --repodir=$HOME/code/whendat --tags="kitware,gitlab,purepy,cv2,gdal"
    python -m xcookie.main --repodir=$HOME/code/whydat --tags="kitware,gitlab,purepy,cv2,gdal"
    python -m xcookie.main --repodir=$HOME/code/howdat --tags="kitware,gitlab,purepy,cv2,gdal"

    # Create this repo
    python -m xcookie.main --repo_name=xcookie --repodir=$HOME/code/xcookie --tags="erotemic,github,purepy"

    # Create a new python repo
    python -m xcookie.main --repo_name=cookiecutter_purepy --repodir=$HOME/code/cookiecutter_purepy --tags="github,purepy"

    # Create a new binary repo
    python -m xcookie.main --repo_name=cookiecutter_binpy --repodir=$HOME/code/cookiecutter_binpy --tags="github,binpy,gdal"

    # Create a new binary gitlab kitware repo
    python -m xcookie.main --repo_name=kwimage_ext --repodir=$HOME/code/kwimage_ext --tags="kitware,gitlab,binpy"

    # Create a new binary github repo
    python -m xcookie.main --repodir=$HOME/code/networkx_algo_common_subtree --tags="github,erotemic,binpy"

    # Create a new purepy github repo
    python -m xcookie.main --repodir=$HOME/code/googledoc --tags="github,erotemic,purepy"

    python -m xcookie.main --repodir=$HOME/code/networkx_algo_common_subtree_cython --tags="github,erotemic,binpy"

    python -m xcookie.main --repo_name=delayed_image --repodir=$HOME/code/delayed_image --tags="kitware,gitlab,purepy,cv2,gdal"


    HOST=https://gitlab.kitware.com
    export PRIVATE_GITLAB_TOKEN=$(git_token_for "$HOST")
    python -m xcookie.main --repo_name=kwutil --repodir=$HOME/code/kwutil --tags="kitware,gitlab,purepy"

    python -m xcookie.main --repo_name=geowatch --repodir=$HOME/code/geowatch --tags="kitware,gitlab,purepy,cv2,gdal"

    python -m xcookie.main --repo_name=stdx --repodir=$HOME/code/stdx --tags="github,purepy,erotemic"

    python -m xcookie.main --repo_name=ustd --repodir=$HOME/code/ustd --tags="github,purepy,erotemic"

    load_secrets
    export PRIVATE_GITLAB_TOKEN=$(git_token_for "https://gitlab.kitware.com")
    python -m xcookie.main --repo_name=simple_dvc --repodir=$HOME/code/simple_dvc --tags="gitlab,kitware,purepy,erotemic"
"""
import toml
import shutil
import ubelt as ub
import tempfile
import scriptconfig as scfg
import xdev
import os
from packaging.version import parse as Version


class SkipFile(Exception):
    pass


# TODO: split up into a configuration that is saved to pyproject.toml and one
# that is on only used when executing
class XCookieConfig(scfg.DataConfig):
    __default__ = {
        'repodir': scfg.Value('.', help='path to the new or existing repo', position=1),

        'repo_name': scfg.Value(None, help='defaults to ``repodir.name``'),
        'mod_name': scfg.Value(None, help='The name of the importable Python module. defaults to ``repo_name``'),
        'pkg_name': scfg.Value(None, help='The name of the installable Python package. defaults to ``mod_name``'),
        'rel_mod_parent_dpath': scfg.Value('.', help=ub.paragraph(
            '''
            The location of the module directory relative to the repository
            root.  This defaults to simply placing the module in ".", but
            another common pattern is to specify this as "./src".
            '''
        )),

        'rotate_secrets': scfg.Value('auto', help='If True will execute secret rotation', isflag=True),
        'refresh_docs': scfg.Value('auto', help='If True will refresh the docs', isflag=True),

        'os': scfg.Value('all', help='all or any of win,osx,linux'),

        'is_new': scfg.Value('auto', help=ub.paragraph(
            '''
            If the repo is detected or specified as being new, then steps to
            create a project for the repo on github/gitlab and other
            initialization procedures will be executed. Otherwise we assume
            that we are updating an existing repo.
            '''
        )),

        'min_python': scfg.Value('3.7'),
        'typed': scfg.Value(None, help='Should be None, False, True, partial or full'),
        'supported_python_versions': scfg.Value('auto', help=ub.paragraph(
            '''
            can specify as a list of explicit major.minor versions. Auto will
            use everything above the min_python version
            ''')),

        'ci_cpython_versions': scfg.Value('auto', help=ub.paragraph(
            '''
            Specify the major.minor CPython versions to use on the CI.
            Will default to the supported_python_versions. E.g. ["3.7", "3.10"]
            ''')),

        'ci_pypy_versions': scfg.Value('auto', help=ub.paragraph(
            '''
            Specify the major.minor PyPy versions to use on the CI.
            Defaults will depend on purepy vs binpy tags.
            ''')),

        'ci_versions_minimal_strict': scfg.Value('min', help='todo: sus out'),
        'ci_versions_full_strict': scfg.Value('max'),
        'ci_versions_minimal_loose': scfg.Value('max'),
        'ci_versions_full_loose': scfg.Value('*'),

        'remote_host': scfg.Value(None, help='if unspecified, attempt to infer from tags'),
        'remote_group': scfg.Value(None, help='if unspecified, attempt to infer from tags'),

        'autostage': scfg.Value(False, help='if true, automatically add changes to version control'),

        'visibility': scfg.Value('public', help='or private. Does limit what we can do'),


        'version': scfg.Value(None, help='repo metadata: url for the project'),
        'url': scfg.Value(None, help='repo metadata: url for the project'),
        'author': scfg.Value(None, help='repo metadata: author for the project'),
        'author_email': scfg.Value(None, help='repo metadata'),
        'description': scfg.Value(None, help='repo metadata'),
        'license': scfg.Value(None, help='repo metadata'),
        'dev_status': scfg.Value('planning'),
        'enable_gpg': scfg.Value(True),
        'defaultbranch': scfg.Value('main'),
        'xdoctest_style': scfg.Value('google', help='type of xdoctest style'),

        'ci_pypi_live_password_varname': scfg.Value('TWINE_PASSWORD', help='variable of the live twine password in your secrets'),
        'ci_pypi_test_password_varname': scfg.Value('TEST_TWINE_PASSWORD', help='variable of the test twine password in your secrets'),

        'regen': scfg.Value(None, help=ub.paragraph(
            '''
            if specified, any modified template file that matches this pattern
            will be considered for re-write
            ''')),

        'tags': scfg.Value('auto', nargs='*', help=ub.paragraph(
            '''
            Tags modify what parts of the template are used.
            Valid tags are:
                "binpy" - do we build binpy wheels?
                "erotemic" - this is an erotemic repo
                "kitware" - this is an kitware repo
                "pyutils" - this is an pyutils repo
                "purepy" - this is a pure python repo
                "gdal" - add in our gdal hack # TODO
                "postgresql" - add in postgresql dependencies
                "cv2" - enable the headless hack
                "notypes" - disable mypy in lint checks
            ''')),

        'interactive': scfg.Value(True),
        'yes': scfg.Value(False, help=ub.paragraph('Say yes to everything')),

        'linter': scfg.Value(True, help=ub.paragraph('if true enables lint checks in CI')),
    }

    def __post_init__(self):
        if self['repodir'] is None:
            self['repodir'] = ub.Path.cwd()
        else:
            self['repodir'] = ub.Path(self['repodir']).absolute()

        try:
            self['repodir'] = find_git_root(self['repodir'])
        except Exception:
            print('assuming the root was given and we are not in a repo yet')

        if self['tags']:
            if isinstance(self['tags'], str):
                self['tags'] = [self['tags']]
            new = []
            for t in self['tags']:
                new.extend([p.strip() for p in t.split(',')])
            self['tags'] = new

        if self['os']:
            if isinstance(self['os'], str):
                self['os'] = [self['os']]
            new = []
            for t in self['os']:
                new.extend([p.strip() for p in t.split(',')])
            self['os'] = set(new)
            if 'all' in self['os']:
                self['os'].add('win')
                self['os'].add('osx')
                self['os'].add('linux')
                self['os'].remove('all')
            self['os'] = sorted(self['os'])

        if self['repo_name'] is None:
            self['repo_name'] = self['repodir'].name
        if self['mod_name'] is None:
            self['mod_name'] = self['repo_name'].replace('-', '_')
        if self['pkg_name'] is None:
            self['pkg_name'] = self['mod_name']
        if self['is_new'] == 'auto':
            self['is_new'] = not (self['repodir'] / '.git').exists()
        if self['rotate_secrets'] == 'auto':
            self['rotate_secrets'] = self['is_new']
        if self['refresh_docs'] == 'auto':
            self['refresh_docs'] = self['is_new']
        if self['author'] is None:
            if 'erotemic' in self['tags']:
                self['author'] = 'Jon Crall'
            else:
                self['author'] = ub.cmd('git config user.name')['out'].strip()
                if self['author'] == 'joncrall':
                    self['author'] = 'Jon Crall'
        if self['license'] is None:
            self['license'] = 'Apache 2'
        if self['author_email'] is None:
            if 'erotemic' in self['tags']:
                self['author_email'] = 'erotemic@gmail.com'
            else:
                self['author_email'] = ub.cmd('git config user.email')['out'].strip()
        if self['version'] is None:
            # TODO: read from __init__.py
            self['version'] = '{mod_dpath}/__init__.py::__version__'
        if self['description'] is None:
            self['description'] = 'The {} module'.format(self['mod_name'])

        if self['supported_python_versions'] == 'auto':
            from xcookie.constants import KNOWN_PYTHON_VERSIONS
            min_python = str(self['min_python'])
            python_versions = [v for v in KNOWN_PYTHON_VERSIONS
                               if Version(v) >= Version(min_python)]
            self['supported_python_versions'] = python_versions

        if self['ci_cpython_versions'] == 'auto':
            self['ci_cpython_versions'] = self['supported_python_versions']

        if self['ci_pypy_versions'] == 'auto':
            if 'purepy' in self['tags']:
                self['ci_pypy_versions'] = ['3.7']
            else:
                self['ci_pypy_versions'] = []

    def _load_pyproject_config(self):
        pyproject_fpath = self['repodir'] / 'pyproject.toml'
        if pyproject_fpath.exists():
            try:
                disk_config = toml.loads(pyproject_fpath.read_text())
            except Exception:
                raise
            return disk_config

    def _load_xcookie_pyproject_settings(self):
        disk_config = self._load_pyproject_config()
        if disk_config is not None:
            settings = disk_config.get('tool', {}).get('xcookie', {})
            return settings

    def confirm(self, msg, default=True):
        """
        Args:
            msg (str): display to the user
            default (bool): default value if non-interactive

        Returns:
            bool:
        """
        if self['interactive']:
            from rich import prompt
            flag = prompt.Confirm.ask(msg)
        else:
            flag = default
        return flag

    def prompt(self, msg, choices, default=True):
        """
        Args:
            msg (str): display to the user
            default (bool): default value if non-interactive

        Returns:
            bool:
        """
        if self['interactive']:
            from xcookie.rich_ext import FuzzyPrompt
            ans = FuzzyPrompt.ask(msg, choices=choices)
        else:
            ans = default
        return ans

    @classmethod
    def main(cls, cmdline=0, **kwargs):
        """
        Main entry point

        Ignore:
            repodir = ub.Path('~/code/pyflann_ibeis').expand()
            kwargs = {
                'repodir': repodir,
                'tags': ['binpy', 'erotemic', 'github'],
            }
            cmdline = 0

        Example:
            repodir = ub.Path.appdir('pypkg/demo/my_new_repo')
            import sys, ubelt
            sys.path.append(ubelt.expandpath('~/code/xcookie'))
            from xcookie.main import *  # NOQA
            kwargs = {
                'repodir': repodir,
            }
            cmdline = 0
        """
        # We load the config multiple times to get the right defaults.
        config = XCookieConfig.cli(cmdline=cmdline, data=kwargs)
        config.normalize()
        settings = config._load_xcookie_pyproject_settings()
        if settings:
            print(f'settings={settings}')
            config = XCookieConfig.cli(cmdline=cmdline, data=kwargs, default=ub.dict_isect(settings, config))
        config.normalize()

        # import xdev
        # xdev.embed()

        import rich
        rich.print('config = {}'.format(ub.repr2(config, nl=1)))
        repodir = ub.Path(config['repodir']).absolute()
        repodir.ensuredir()

        self = TemplateApplier(config)
        self.setup()
        self.apply()


class TemplateApplier:
    """
    The primary xcookie autogeneration class.

    Note:
        this does not write any files unless you call setup (which just writes
        to a temporary directory) or apply (which can destructively clobber
        things).
    """

    def __init__(self, config):
        if isinstance(config, dict):
            config = XCookieConfig(**config)

        self.config = config
        self.repodir = ub.Path(self.config['repodir'])
        if self.config['repo_name'] is None:
            self.config['repo_name'] = self.repodir.name
        self.repo_name = self.config['repo_name']
        self._tmpdir = tempfile.TemporaryDirectory(prefix=self.repo_name)

        self.template_infos = None
        try:
            xcookie_dpath = ub.Path(__file__).parent.parent
        except NameError:
            xcookie_dpath = ub.Path('~/misc/templates/xcookie').expand()
        self.template_dpath = xcookie_dpath
        self.staging_dpath = ub.Path(self._tmpdir.name)
        self.remote_info = {
            'type': 'unknown'
        }

    def apply(self):
        """
        Does the actual modification of the target repo.

        Has special logic to handle building new respos versus updating repos.
        """
        self.vcs_checks()
        self.copy_staged_files()
        if self.config['refresh_docs']:
            self.refresh_docs()
        if self.config['rotate_secrets']:
            self.rotate_secrets()
        self.print_help_tips()

        if self.config['autostage']:
            self.autostage()

    def autostage(self):
        import git
        repo = git.Repo(self.repodir)

        # Find untracked files
        untracked = []
        for info in self.staging_infos:
            fpath = info['repo_fpath']
            if not repo.git.ls_files(fpath):
                untracked.append(fpath)

        repo.git.add(untracked)

    @property
    def has_git(self):
        return (self.config['repodir'] / '.git').exists()

    @property
    def rel_mod_dpath(self):
        return ub.Path(self.config['rel_mod_parent_dpath']) / self.config['mod_name']

    @property
    def mod_dpath(self):
        return self.repodir / self.rel_mod_dpath

    @property
    def mod_name(self):
        return self.config['mod_name']

    @property
    def pkg_name(self):
        return self.config['pkg_name']

    def _build_template_registry(self):
        """
        Take stock of the files in the template repo and ensure they all have
        appropriate properties.
        """
        from xcookie import rc

        rel_mod_dpath = self.rel_mod_dpath

        self.template_infos = [
            # {'template': 1, 'overwrite': False, 'fname': '.circleci/config.yml'},
            # {'template': 1, 'overwrite': False, 'fname': '.travis.yml'},

            {'template': 0, 'overwrite': 1, 'fname': 'dev/setup_secrets.sh', 'enabled': self.config['enable_gpg']},

            {'template': 0, 'overwrite': 0, 'fname': '.gitignore'},
            # {'template': 1, 'overwrite': 1, 'fname': '.coveragerc'},
            {'template': 1, 'overwrite': 1, 'fname': '.readthedocs.yml',
             'dynamic': 'build_readthedocs' },
            # {'template': 0, 'overwrite': 1, 'fname': 'pytest.ini'},

            {'template': 0, 'overwrite': 0, 'fname': 'pyproject.toml',
             'dynamic': 'build_pyproject'},

            {'template': 1, 'overwrite': 0, 'fname': 'setup.py',
             # 'input_fname': rc.resource_fpath('setup.py.in'),
             'dynamic': 'build_setup',
             'perms': 'x'},

            {'template': 0, 'overwrite': 0, 'fname': 'docs/source/index.rst',
             'dynamic': 'build_docs_index'},
            {'template': 0, 'overwrite': 0, 'fname': 'README.rst',
             'dynamic': 'build_readme'},
            #
            {'source': 'dynamic', 'overwrite': 0, 'fname': 'CHANGELOG.md'},
            {'source': 'dynamic', 'overwrite': 0, 'fname': rel_mod_dpath / '__init__.py'},
            # {'source': 'dynamic', 'overwrite': 0, 'fname': rel_mod_dpath / '__main__.py'},
            {'source': 'dynamic', 'overwrite': 0, 'fname': 'tests/test_import.py'},

            {'template': 0, 'overwrite': 1, 'fname': '.github/dependabot.yml', 'tags': 'github'},

            # {'template': 0, 'overwrite': 1,
            #  'tags': 'binpy,github',
            #  'fname': '.github/workflows/test_binaries.yml',
            #  'input_fname': rc.resource_fpath('test_binaries.yml.in')},

            {'template': 1, 'overwrite': 1,
             'tags': 'github',
             'fname': '.github/workflows/tests.yml',
             'dynamic': 'build_github_actions',
             # 'input_fname': rc.resource_fpath('tests.yml.in')
             },

            {'template': 0, 'overwrite': 1, 'fname': '.gitlab-ci.yml', 'tags': 'gitlab,purepy',
             # 'input_fname': rc.resource_fpath('gitlab-ci.purepy.yml.in')
             'dynamic': 'build_gitlab_ci'
             },

            {'template': 0, 'overwrite': 1, 'fname': '.rules.yml', 'tags': 'gitlab',
             # 'input_fname': rc.resource_fpath('gitlab-ci.purepy.yml.in')
             'dynamic': 'build_gitlab_rules'
             },

            {'template': 0, 'overwrite': 1, 'fname': '.gitlab-ci.yml', 'tags': 'gitlab,binpy',
             'input_fname': rc.resource_fpath('gitlab-ci.binpy.yml.in')},
            # {'template': 1, 'overwrite': False, 'fname': 'appveyor.yml'},

            {'template': 1, 'overwrite': 0, 'fname': 'CMakeLists.txt',
             'tags': 'binpy',
             'input_fname': rc.resource_fpath('CMakeLists.txt.in')},

            # {'template': 0, 'overwrite': 1, 'fname': 'dev/make_strict_req.sh', 'perms': 'x'},

            {'template': 0, 'overwrite': 1, 'fname': 'requirements.txt',  'dynamic': 'build_requirements'},
            {'template': 0, 'overwrite': 1, 'fname': 'requirements/graphics.txt', 'tags': 'cv2'},
            {'template': 0, 'overwrite': 1, 'fname': 'requirements/headless.txt', 'tags': 'cv2'},
            {'template': 0, 'overwrite': 1, 'fname': 'requirements/gdal.txt', 'tags': 'gdal'},
            {'template': 0, 'overwrite': 1, 'fname': 'requirements/gdal-strict.txt', 'tags': 'gdal'},
            {'template': 0, 'overwrite': 0, 'fname': 'requirements/optional.txt'},
            {'template': 0, 'overwrite': 0, 'fname': 'requirements/runtime.txt'},
            {'template': 0, 'overwrite': 0, 'fname': 'requirements/tests.txt'},
            {'template': 0, 'overwrite': 0, 'fname': 'requirements/docs.txt'},
            {'template': 1, 'overwrite': 1, 'fname': 'docs/source/conf.py', 'dynamic': 'build_docs_conf'},
            {'template': 1, 'overwrite': 1, 'fname': 'docs/Makefile'},
            {'template': 1, 'overwrite': 1, 'fname': 'docs/make.bat'},

            # {'template': 0, 'overwrite': 0, 'fname': 'docs/source/_static', 'path_type': 'dir'},
            # {'template': 0, 'overwrite': 0, 'fname': 'docs/source/_templates', 'path_type': 'dir'},

            {'template': 0, 'overwrite': 1, 'fname': 'publish.sh', 'perms': 'x'},
            {'template': 1, 'overwrite': 1, 'fname': 'build_wheels.sh', 'perms': 'x', 'tags': 'binpy'},
            {'template': 1, 'overwrite': 1, 'fname': 'run_doctests.sh', 'perms': 'x',
             'dynamic': 'build_run_doctests',
             },  # TODO: template with xdoctest-style

            {'template': 0, 'overwrite': 1, 'fname': 'run_linter.sh', 'perms': 'x',
             'dynamic': 'build_run_linter'},

            # TODO: template a clean script
            {'template': 1, 'overwrite': 1, 'fname': 'run_tests.py',
             'perms': 'x', 'tags': 'binpy',
             'input_fname': rc.resource_fpath('run_tests.binpy.py.in')},

            {'template': 1, 'overwrite': 1, 'fname': 'run_tests.py',
             'perms': 'x', 'tags': 'purepy',
             'input_fname': rc.resource_fpath('run_tests.purepy.py.in')},

        ]
        if 0:
            # Checker and help autopopulate
            template_contents = []
            dname_blocklist = {
                '__pycache__',
                'old',
                '.circleci',
                'xcookie',
                '.git',
            }
            fname_blocklist = set()
            for root, ds, fs in self.template_dpath.walk():
                for d in set(ds) & dname_blocklist:
                    ds.remove(d)
                fs = set(fs) - fname_blocklist
                if fs:
                    rel_root = root.relative_to(self.template_dpath)
                    for fname in fs:
                        abs_fpath = root / fname
                        if abs_fpath.name.endswith('.in'):
                            is_template = 1
                        else:
                            try:
                                is_template = int('xcookie' in abs_fpath.read_text())
                            except Exception:
                                is_template = 0
                            # is_template = 0
                        rel_fpath = rel_root / fname
                        # overwrite indicates if we dont expect the user to
                        # make modifications
                        template_contents.append({
                            'template': is_template,
                            'overwrite': False,
                            'fname': os.fspath(rel_fpath),
                        })
            print('template_contents = {}'.format(ub.repr2(sorted(template_contents, key=lambda x: x['fname']), nl=1, sort=0)))
            known_fpaths = {d['fname'] for d in self.template_infos}
            exist_fpaths = {d['fname'] for d in template_contents}
            unexpected_fpaths = exist_fpaths - known_fpaths
            if unexpected_fpaths:
                print(f'WARNING UNREGISTERED unexpected_fpaths={unexpected_fpaths}')

    @property
    def tags(self):
        return set(self.config['tags'])

    def _presetup(self):
        """
        This logic used to live in setup, but it doesn't have external side
        effects, so it would be good if we were able have these fields
        populated on initialization for tests.
        """
        tags = set(self.config['tags'])
        self.remote_info = {
            'type': 'unknown'
        }

        if self.config.url is None and 'github' not in tags and 'gitlab' not in tags:
            # We can infer this if the repo already exists.
            git_dpath = self.repodir / '.git'
            if git_dpath.exists():
                remote_url = ub.cmd('git remote -v  get-url origin', cwd=self.repodir)['out'].strip()
                if self.config.url is None:
                    self.config.url = remote_url

        if self.config['remote_host'] is not None:
            self.remote_info['host'] = self.config['remote_host']

        if self.config['remote_group'] is not None:
            self.remote_info['group'] = self.config['remote_group']

        url = self.config.get('url', None)
        if url is not None:
            info = _parse_remote_url(url)
            self.remote_info['group'] = info['group']
            self.remote_info['host'] = info['host']
            self.remote_info['repo_name'] = info['repo_name']
            if 'github' in self.remote_info['host']:
                self.remote_info['type'] = 'github'
            if 'gitlab' in self.remote_info['host']:
                self.remote_info['type'] = 'gitlab'

        if 'gitlab' in tags:
            self.remote_info['type'] = 'gitlab'
        if 'github' in tags:
            self.remote_info['type'] = 'github'

        default_remote_info = {}

        if self.remote_info['type'] == 'gitlab':
            if 'kitware' in tags:
                default_remote_info['host'] = 'https://gitlab.kitware.com'
                default_remote_info['group'] = 'computer-vision'  # hack
        if self.remote_info['type'] == 'github':
            default_remote_info['host'] = 'https://github.com'
            if 'erotemic' in tags:
                default_remote_info['group'] = 'Erotemic'  # hack
            if 'pyutils' in tags:
                default_remote_info['group'] = 'pyutils'  # hack

        self.remote_info = ub.udict(default_remote_info) | ub.udict(self.remote_info)

        self.remote_info['repo_name'] = self.config['repo_name']

        if 'group' in self.remote_info and 'host' in self.remote_info:
            self.config['remote_host'] = self.remote_info['host']
            self.config['remote_group'] = self.remote_info['group']
            self.remote_info['url'] = '/'.join([self.remote_info['host'], self.remote_info['group'], self.config['repo_name']])
            self.remote_info['git_url'] = '/'.join([self.remote_info['host'], self.remote_info['group'], self.config['repo_name'] + '.git'])

    def setup(self):
        """
        Finalizes a few variables and writes the "clean" template to the
        staging directory.
        """
        self._presetup()

        tags = set(self.config['tags'])

        if self.remote_info['type'] == 'unknown':
            print(f'tags={tags}')
            print('self.remote_info = {}'.format(ub.repr2(self.remote_info, nl=1)))
            raise Exception('Specify github or gitlab in tags')

        if 'group' not in self.remote_info:
            print(f'tags={tags}')
            print('self.remote_info = {}'.format(ub.repr2(self.remote_info, nl=1)))
            raise Exception('Unknown user / group, specify a tag for a known user. Or a URL in the pyproject.toml [tool.xcookie]')

        self._build_template_registry()
        self.stage_files()
        return self

    def copy_staged_files(self):
        stats, tasks = self.gather_tasks()
        copy_tasks = tasks['copy']
        perm_tasks = tasks['perms']
        mkdir_tasks = tasks['mkdir']
        task_summary = ub.map_vals(len, tasks)
        if any(task_summary.values()):
            print('task_summary = {}'.format(ub.repr2(task_summary, nl=1)))
            ans = self.config.prompt('What parts of the patch to apply?', ['yes', 'all', 'some', 'none'])
            if ans in {'all', 'yes'}:
                dirs = {d.parent for s, d in copy_tasks}
                for d in dirs:
                    d.ensuredir()
                for d in mkdir_tasks:
                    d.ensuredir()
                for src, dst in copy_tasks:
                    shutil.copy2(src, dst)
                for fname, mode in perm_tasks:
                    os.chmod(fname, mode)
            elif ans == 'some':
                dirs = {d.parent for s, d in copy_tasks}
                for d in dirs:
                    d.ensuredir()
                for d in mkdir_tasks:
                    d.ensuredir()
                for src, dst in copy_tasks:
                    if self.config.confirm(f'Apply {dst}?'):
                        shutil.copy2(src, dst)
                for fname, mode in perm_tasks:
                    os.chmod(fname, mode)

    def vcs_checks(self):
        # repodir = self.config['repodir']
        # mod_dpath = None
        # if mod_dpath is None:
        #     mod_name = self.config['mod_name']
        #     mod_dpath = repodir / mod_name

        # package_structure = [
        #     repodir / 'CHANGELOG.md',
        #     mod_dpath / '__init__.py',
        #     mod_dpath / '__main__.py',
        # ]
        # missing = []
        # for fpath in package_structure:
        #     if not fpath.exists():
        #         missing.append(fpath)
        # if missing:
        #     print('missing = {}'.format(ub.repr2(missing, nl=1)))
        if self.config['is_new']:
            create_new_repo_info = ub.codeblock(
                f'''
                TODO: call the APIS
                git init
                gh repo create {self.repo_name} --public
                # https://cli.github.com/manual/gh_repo_create
                ''')
            print(create_new_repo_info)
            import cmd_queue
            queue = cmd_queue.Queue.create(cwd=self.repodir)
            git_dpath = self.repodir / '.git'
            if not git_dpath.exists():
                queue.submit('git init')
                queue.sync().submit(f'git remote add origin {self.remote_info["url"]}')

            if 'erotemic' in self.tags:
                # TODO: ensure this works
                # for erotemic repos, configure the local user and email
                # TODO: make an xcookie user configuration where this
                # information is pulled from.
                queue.sync().submit('git config --local user.name "Jon Crall"')
                queue.sync().submit('git config --local user.email "erotemic@gmail.com"')
                # see also:
                # ~/local/scripts/git-autoconf-gpgsign.sh Erotemic
                queue.sync().submit('git config --local commit.gpgsign true')
                queue.sync().submit('git config --local user.signingkey 4AC8B478335ED6ED667715F3622BE571405441B4')

            if queue.jobs:
                queue.rprint()
                if self.config.confirm('Do git init?'):
                    queue.run()

            if self.config.confirm('Do you want to create the repo on the remote?'):
                if 'gitlab' in self.tags:
                    """
                    Requires user do something to load secrets:

                    load_secrets
                    HOST=https://gitlab.kitware.com
                    export PRIVATE_GITLAB_TOKEN=$(git_token_for "$HOST")
                    """
                    from xcookie.vcs_remotes import GitlabRemote
                    vcs_remote = GitlabRemote(
                        proj_name=self.remote_info['repo_name'],
                        proj_group=self.remote_info['group'],
                        url=self.remote_info['host'],
                        visibility=self.config['visibility'])
                    vcs_remote.auth()
                    vcs_remote.new_project()
                elif 'github' in self.tags:
                    from xcookie.vcs_remotes import GithubRemote
                    vcs_remote = GithubRemote(self.remote_info['repo_name'])
                    vcs_remote.new_project()
                else:
                    raise NotImplementedError('unknown vcs remote')

    def _stage_file(self, info):
        """
        Write a single file to the staging directory based on its template
        info.

        Example:
            >>> from xcookie.main import *  # NOQA
            >>> dpath = ub.Path.appdir('xcookie/tests/test-stage').delete().ensuredir()
            >>> kwargs = {
            >>>     'repodir': dpath / 'testrepo',
            >>>     'tags': ['gitlab', 'kitware', 'purepy', 'cv2'],
            >>>     'rotate_secrets': False,
            >>>     'is_new': False,
            >>>     'interactive': False,
            >>> }
            >>> config = XCookieConfig.cli(cmdline=0, data=kwargs)
            >>> config.normalize()
            >>> print('config = {}'.format(ub.repr2(dict(config), nl=1)))
            >>> self = TemplateApplier(config)
            >>> self._build_template_registry()
            >>> info = [d for d in self.template_infos if d['fname'] == 'setup.py'][0]
            >>> info = [d for d in self.template_infos if d['fname'] == '.gitlab-ci.yml'][0]
            >>> self._stage_file(info)
        """
        # print('info = {!r}'.format(info))
        tags = info.get('tags', None)
        if tags:
            tags = set(tags.split(','))
            if not set(self.config['tags']).issuperset(tags):
                raise SkipFile

        path_name = info['fname']
        path_type = info.get('path_type', 'file')

        stage_fpath = self.staging_dpath / path_name
        info['stage_fpath'] = stage_fpath
        info['repo_fpath'] = self.repodir / path_name
        info['path_type'] = path_type
        if path_type == 'dir':
            stage_fpath.ensuredir()
        else:
            stage_fpath.parent.ensuredir()
            dynamic = info.get('dynamic', '') or info.get('source', '') == 'dynamic'
            if dynamic:
                dynamic_var = info.get('dynamic', '')
                if dynamic_var == '':
                    text = self.lut(info)
                else:
                    text = getattr(self, dynamic_var)()
                try:
                    stage_fpath.write_text(text)
                except Exception:
                    print(f'text={text}')
                    raise
                    pass
            else:
                in_fname = info.get('input_fname', path_name)
                raw_fpath = self.template_dpath / in_fname
                if not raw_fpath.exists():
                    raise IOError(f'Template file: raw_fpath={raw_fpath} does not exist')
                shutil.copy2(raw_fpath, stage_fpath)

                self._apply_xcookie_directives(stage_fpath)

                if info['template']:
                    xdev.sedfile(stage_fpath, 'xcookie', self.repo_name, verbose=0)
                    xdev.sedfile(stage_fpath, '<mod_name>', self.mod_name, verbose=0)
                    xdev.sedfile(stage_fpath, '<rel_mod_dpath>', str(self.rel_mod_dpath), verbose=0)
                    # FIXME: use configuration from pyproject.toml
                    author = ub.cmd('git config --global user.name')['out'].strip()
                    author_email = ub.cmd('git config --global user.email')['out'].strip()
                    xdev.sedfile(stage_fpath, '<AUTHOR>', author, verbose=0)
                    xdev.sedfile(stage_fpath, '<AUTHOR_EMAIL>', author_email, verbose=0)
        return info

    def _apply_xcookie_directives(self, stage_fpath):
        text = stage_fpath.read_text()
        from xcookie.directive import DirectiveExtractor
        namespace = 'xcookie'
        commands = ['UNCOMMENT_IF', 'COMMENT_IF']
        extractor = DirectiveExtractor(namespace, commands)

        import re
        def comment_line(line):
            """

            line = '       #- pip install .[tests-strict,headless-strict]  # testrepo: +UNCOMMENT_IF(cv2)'
            uncomment_line(line)

            cases = [
                '   foobar',
                'foobar',
                '   def fds(): # hello',
            ]
            for line in cases:
                cline = comment_line(line)
                uline = uncomment_line(cline)
                print(f'line={line}')
                print(f'cline={cline}')
                print(f'uline={uline}')
                assert uline == line
            cases = [
                '#   foobar',
                '   #  foobar',
                '   #foobar',
                '#foobar',
                '#  foobar',
                '# foobar',
                '#   def fds(): # hello',
                '   #def fds(): # hello',
            ]
            for line in cases:
                uline = uncomment_line(cline)
                cline = comment_line(line)
                print(f'line={line}')
                print(f'cline={cline}')
                print(f'uline={uline}')
            """
            return re.sub(r'^(\s*)([^\s])', r'\g<1># \g<2>', line)

        def uncomment_line(line):
            return re.sub(r'^(\s*)#\s*', r'\g<1>', line, count=1)

        def tags_satisfied(directive, tags):
            value = tags.issuperset(set(directive.args))
            return value

        tags = set(self.config['tags'])
        new_lines = []
        did_work = 0
        for line in text.split('\n'):
            extracted = list(extractor.extract(line))
            # if 'COMMENT' in line:
            #     print(f'line={line}')
            #     print(f'extracted={extracted}')
            if extracted:
                for directive in extracted:
                    action = None
                    if directive.name == 'COMMENT_IF':
                        value = tags_satisfied(directive, tags)
                        if value:
                            action = comment_line
                            # print(f'action={action}')
                            did_work = 1
                    if directive.name == 'UNCOMMENT_IF':
                        value = tags_satisfied(directive, tags)
                        if value:
                            action = uncomment_line
                            print(f'action={action}')
                            did_work = 1
                    if action is not None:
                        # print(f'directive.name={directive.name}')
                        # print(f'action={action}')
                        # print(f'old line={line}')
                        line = action(line)
                        # print(f'new line={line}')
            new_lines.append(line)

        if did_work:
            stage_fpath.write_text('\n'.join(new_lines))

    def stage_files(self):
        self.staging_infos = []
        for info in ub.ProgIter(self.template_infos, desc='staging'):
            if not info.get('enabled', True):
                continue
            try:
                info = self._stage_file(info)
            except SkipFile:
                continue
            else:
                self.staging_infos.append(info)

        if 1:
            import pandas as pd
            # print('self.staging_infos = {}'.format(ub.repr2(self.staging_infos, nl=1)))
            df = pd.DataFrame(self.staging_infos)
            print(df)

    def gather_tasks(self):
        tasks = {
            'copy': [],
            'perms': [],
            'mkdir': [],
        }
        stats = {
            'missing': [],
            'modified': [],
            'dirty': [],
            'clean': [],
            'missing_dir': [],
        }

        if self.config['regen'] is not None:
            regen_pat = xdev.Pattern.coerce(self.config['regen'])
        else:
            regen_pat = None

        for info in self.staging_infos:
            stage_fpath = info['stage_fpath']
            repo_fpath = info['repo_fpath']
            if not repo_fpath.exists():
                if stage_fpath.is_dir():
                    tasks['mkdir'].append(repo_fpath)
                    stats['missing_dir'].append(repo_fpath)
                else:
                    stats['missing'].append(repo_fpath)
                    tasks['copy'].append((stage_fpath, repo_fpath))
                    stage_text = stage_fpath.read_text()
                    difftext = xdev.difftext('', stage_text[:1000], colored=1)
                    print(f'<NEW FPATH={repo_fpath}>')
                    print(difftext)
                    print(f'<END FPATH={repo_fpath}>')
            else:
                assert stage_fpath.exists()
                if stage_fpath.is_dir():
                    continue
                repo_text = repo_fpath.read_text()
                stage_text = stage_fpath.read_text()
                if stage_text.strip() == repo_text.strip():
                    difftext = None
                else:
                    difftext = xdev.difftext(repo_text, stage_text, colored=1)
                if difftext:
                    want_rewrite = info['overwrite']
                    if not want_rewrite:
                        if regen_pat is not None:
                            if regen_pat.search(info['fname']):
                                want_rewrite = True

                    if want_rewrite:
                        tasks['copy'].append((stage_fpath, repo_fpath))
                        stats['dirty'].append(repo_fpath)
                        print(f'<DIFF FOR repo_fpath={repo_fpath}>')
                        print(difftext)
                        print(f'<END DIFF repo_fpath={repo_fpath}>')
                    else:
                        stats['modified'].append(repo_fpath)
                else:
                    stats['clean'].append(repo_fpath)

            if 'x' in info.get('perms', ''):
                import stat
                if info['repo_fpath'].exists():
                    st = ub.Path(info['repo_fpath']).stat()
                    mode_want = st.st_mode | stat.S_IEXEC
                    if mode_want != st.st_mode:
                        tasks['perms'].append((info['repo_fpath'], mode_want))
                # else:
                #     tasks['perms'].append((info['repo_fpath'], mode_want))

        print('stats = {}'.format(ub.repr2(stats, nl=2)))
        return stats, tasks

    def build_requirements(self):
        # existing = (self.repodir / 'requirements').ls()
        candidate_all_requirements = [
            'requirements/runtime.txt',
            'requirements/tests.txt',
            'requirements/optional.txt',
            'requirements/build.txt',
            'requirements/postgresql.txt',
        ]
        requirement_lines = []
        for fpath_rel in candidate_all_requirements:
            fpath_rel = ub.Path(fpath_rel)
            fpath = self.repodir / fpath_rel
            if fpath.exists():
                requirement_lines.append('-r ' + os.fspath(fpath_rel))

        text = '\n'.join(requirement_lines)
        return text

    def refresh_docs(self):
        docs_dpath = self.repodir / 'docs'
        docs_source_dpath = (docs_dpath / 'source').ensuredir()
        command = f'sphinx-apidoc -f -o "{docs_source_dpath}" "{self.mod_dpath}" --separate'
        ub.cmd(command, verbose=3, check=True, cwd=docs_dpath)
        if self.has_git:
            ub.cmd(f'git add {docs_source_dpath}/*.rst', verbose=3, check=True, cwd=docs_dpath)
            # ub.cmd('make html', verbose=3, check=True, cwd=docs_dpath)

    def rotate_secrets(self):
        setup_secrets_fpath = self.repodir / 'dev/setup_secrets.sh'
        # dev/public_gpg_key
        # if self.config.confirm('Ready to rotate secrets?'):
        if 'erotemic' in self.config['tags']:
            environ_export = 'setup_package_environs_github_erotemic'
            upload_secret_cmd = 'upload_github_secrets'
        elif 'pyutils' in self.config['tags']:
            environ_export = 'setup_package_environs_github_pyutils'
            upload_secret_cmd = 'upload_github_secrets'
        elif 'kitware' in self.config['tags']:
            environ_export = 'setup_package_environs_gitlab_kitware'
            upload_secret_cmd = 'upload_gitlab_repo_secrets'
        else:
            raise Exception

        import cmd_queue
        script = cmd_queue.Queue.create(cwd=self.repodir, backend='serial',
                                        log=False)
        script.submit(f'source {setup_secrets_fpath}', log=False)
        script.sync().submit(f'{environ_export}', log=False)
        script.sync().submit('source $(secret_loader.sh)', log=False)
        script.sync().submit('export_encrypted_code_signing_keys', log=False)
        # script.sync().submit('git commit -am "Updated secrets"')
        script.sync().submit(f'{upload_secret_cmd}', log=False)
        # script.submit(ub.codeblock(
        #     f'''
        #     cd {self.repodir}
        #     source {setup_secrets_fpath}
        #     {environ_export}
        #     load_secrets
        #     export_encrypted_code_signing_keys
        #     git commit -am "Updated secrets"
        #     {upload_secret_cmd}
        #     '''))
        script.rprint()
        # print('FIXME: for now, you need to manually execute this')
        # print('Note: need to load_secrets before running this')
        if self.config.confirm('Ready to rotate secrets?'):
            script.run(system=True, mode='bash')
            # script.run(system=True)

    def print_help_tips(self):
        text = ub.codeblock(
            f'''
            Things that xcookie might eventually do that you should do for
            yourself for now:

            * Add typing to the module

                # xdev requires the non-binary mypy
                pip install -U mypy --no-binary :all:

                # Generate stubs and check them
                xdev docstubs ./{self.rel_mod_dpath} && mypy ./{self.rel_mod_dpath}

                # Then make sure you have typed = true in the [tool.xcookie]
                # section of pyproject.toml and regenerate setup.py

            ''')
        print(text)

    def build_readthedocs(self):
        """
        Returns:
            str: templated code
        """
        from xcookie.builders import readthedocs
        return readthedocs.build_readthedocs(self)

    def build_setup(self):
        """
        Returns:
            str: templated code
        """
        from xcookie.builders import setup
        return setup.build_setup(self)

    def build_pyproject(self):
        """
        Returns:
            str: templated code
        """
        from xcookie.builders import pyproject
        return pyproject.build_pyproject(self)

    def build_github_actions(self):
        from xcookie.builders import github_actions
        return github_actions.build_github_actions(self)

    def build_gitlab_ci(self):
        from xcookie.builders import gitlab_ci
        return gitlab_ci.build_gitlab_ci(self)

    def build_run_linter(self):
        text = ub.codeblock(
            f"""
            #!/bin/bash
            flake8 --count --select=E9,F63,F7,F82 --show-source --statistics {self.rel_mod_dpath}
            flake8 --count --select=E9,F63,F7,F82 --show-source --statistics ./tests
            """
        )
        return text

    def build_gitlab_rules(self):
        from xcookie.builders import gitlab_ci
        return gitlab_ci.build_gitlab_rules(self)

    def build_readme(self):
        from xcookie.builders import readme
        return readme.build_readme(self)

    def build_docs_index(self):
        from xcookie.builders import docs
        return docs.build_docs_index(self)

    def build_docs_conf(self):
        from xcookie.builders import docs_conf
        return docs_conf.build_docs_conf(self)

    def build_run_doctests(self):
        return ub.codeblock(
            f'''
            #!/usr/bin/env bash
            xdoctest {self.rel_mod_dpath} --style={self.config['xdoctest_style']} all "$@"
            ''')

    def lut(self, info):
        """
        Returns:
            str: templated code
        """
        fname = ub.Path(info['fname']).name
        if fname == 'CHANGELOG.md':
            return ub.codeblock(
                '''
                # Changelog
                We [keep a changelog](https://keepachangelog.com/en/1.0.0/).
                We aim to adhere to [semantic versioning](https://semver.org/spec/v2.0.0.html).

                ## [Version 0.0.1] -

                ### Added
                * Initial version
                ''')
        elif fname == 'test_import.py':
            return ub.codeblock(
                f'''
                def test_import():
                    import {self.config['mod_name']}
                ''')
        elif fname == '__main__.py':
            return ub.codeblock(
                '''
                #!/usr/bin/env python
                ''')
        elif fname == '__init__.py':
            return ub.codeblock(
                f'''
                """
                Basic
                """
                __version__ = '{self.config['version']}'
                __author__ = '{self.config['author']}'
                __author_email__ = '{self.config['author_email']}'
                __url__ = '{self.config['url']}'

                __mkinit__ = """
                mkinit {info['repo_fpath']}
                """
                ''')
        else:
            raise KeyError(fname)

    def _docs_quickstart():
        # Probably just need to copy/paste the conf.py
        r"""

        TODO:
            - [ ] Auto-edit the index.py to include the magic reference to `__init__`
            - [ ] If this is a utility library, populate the "usefulness section"
            - [ ] Try and find out how to auto-expand the toc tree
            - [ ] Auto-run the sphinx-apidoc for the user

        REPO_NAME=xcookie
        REPO_DPATH=$HOME/code/$REPO_NAME
        AUTHOR=$(git config --global user.name)
        cd $REPO_DPATH/docs
        sphinx-quickstart -q --sep \
            --project="$REPO_NAME" \
            --author="$AUTHOR" \
            --ext-autodoc \
            --ext-viewcode \
            --ext-intersphinx \
            --ext-todo \
            --extensions=sphinx.ext.autodoc,sphinx.ext.viewcode,sphinx.ext.napoleon,sphinx.ext.intersphinx,sphinx.ext.todo,sphinx.ext.autosummary \
            "$REPO_DPATH/docs"

        # THEN NEED TO:
        REPO_NAME=kwarray
        REPO_DPATH=$HOME/code/$REPO_NAME
        MOD_DPATH=$REPO_DPATH/$REPO_NAME
        sphinx-apidoc -f -o "$REPO_DPATH/docs/source"  "$MOD_DPATH" --separate
        cd "$REPO_DPATH/docs"
        make html

        The user will need to enable the repo on their readthedocs account:
        https://readthedocs.org/dashboard/import/manual/?

        To enable the read-the-docs go to https://readthedocs.org/dashboard/ and login

        Make sure you have a .readthedocs.yml file

        Click import project: (for github you can select, but gitlab you need to import manually)
            Set the Repository NAME: $REPO_NAME
            Set the Repository URL: $REPO_URL

        For gitlab you also need to setup an integrations and add gitlab
        incoming webhook Then go to $REPO_URL/hooks and add the URL

        Will also need to activate the main branch:
            https://readthedocs.org/projects/xcookie/versions/
        """
        pass


def main():
    XCookieConfig.main(cmdline={
        'strict': True,
        'autocomplete': True,
    })


def _parse_remote_url(url):
    info = {}
    if url.startswith('https://'):
        parts = url.split('https://')[1].split('/')
        info['host'] = 'https://' + parts[0]
        info['group'] = parts[1]
        info['repo_name'] = parts[2]
    elif url.startswith('git@'):
        url.split('git@')[1]
        parts = url.split('git@')[1].split(':')
        info['host'] = 'https://' + parts[0]
        info['group'] = parts[1].split('/')[0]
        info['repo_name'] = parts[1].split('/')[1]
    else:
        raise ValueError(url)
    return info


def find_git_root(dpath):
    cwd = ub.Path(dpath).resolve()
    parts = cwd.parts
    found = None
    for i in reversed(range(0, len(parts) + 1)):
        p = ub.Path(*parts[0:i])
        cand = p / '.git'
        if cand.exists():
            found = p
            break
    if found is None:
        raise Exception('cannot find git root')
    return found

if __name__ == '__main__':
    """
    CommandLine:
        python ~/misc/templates/xcookie/apply_template.py --help
    """
    main()



def build_docs_conf(self):
    import ubelt as ub
    import datetime

    author = self.config['author']
    if isinstance(author, list):
        author_str = ' '.join(author)
    else:
        author_str = author

    fmtkw = {
        'repo_name': self.config['repo_name'],
        'repo_dashname': self.config['repo_name'].replace('_', '-'),
        'mod_name': self.config['mod_name'],
        'rel_mod_dpath': str(self.rel_mod_dpath),
        'repo_url': self.remote_info['url'],
        'author': author_str,
        'year': datetime.datetime.now().year,
        'repodir_wrt_home': self.repodir.shrinkuser()
    }

    text = ub.codeblock(
        r'''
        """
        Notes:
            Based on template code in:
                ~/code/xcookie/xcookie/builders/docs_conf.py
                ~/code/xcookie/xcookie/rc/conf_ext.py

            http://docs.readthedocs.io/en/latest/getting_started.html

            pip install sphinx sphinx-autobuild sphinx_rtd_theme sphinxcontrib-napoleon

            cd {repodir_wrt_home}
            mkdir -p docs
            cd docs

            sphinx-quickstart

            # need to edit the conf.py

            cd {repodir_wrt_home}/docs
            sphinx-apidoc --private -f -o {repodir_wrt_home}/docs/source {repodir_wrt_home}/{rel_mod_dpath} --separate
            make html

            git add source/*.rst

            Also:
                To turn on PR checks

                https://docs.readthedocs.io/en/stable/guides/autobuild-docs-for-pull-requests.html

                https://readthedocs.org/dashboard/{repo_dashname}/advanced/

                ensure your github account is connected to readthedocs
                https://readthedocs.org/accounts/social/connections/

                ### For gitlab

                The user will need to enable the repo on their readthedocs account:
                https://readthedocs.org/dashboard/import/manual/?

                To enable the read-the-docs go to https://readthedocs.org/dashboard/ and login

                Make sure you have a .readthedocs.yml file

                Click import project: (for github you can select, but gitlab you need to import manually)
                    Set the Repository NAME: {repo_name}
                    Set the Repository URL: {repo_url}

                For gitlab you also need to setup an integrations and add gitlab
                incoming webhook

                    https://readthedocs.org/dashboard/{repo_dashname}/integrations/create/

                Then go to

                    {repo_url}/hooks

                and add the URL

                select push, tag, and merge request

                See Docs for more details https://docs.readthedocs.io/en/stable/integrations.html

                Will also need to activate the main branch:
                    https://readthedocs.org/projects/{repo_dashname}/versions/
        """
        #
        # Configuration file for the Sphinx documentation builder.
        #
        # This file does only contain a selection of the most common options. For a
        # full list see the documentation:
        # http://www.sphinx-doc.org/en/stable/config

        # -- Path setup --------------------------------------------------------------

        # If extensions (or modules to document with autodoc) are in another directory,
        # add these directories to sys.path here. If the directory is relative to the
        # documentation root, use os.path.abspath to make it absolute, like shown here.
        #
        # import os
        # import sys
        # sys.path.insert(0, os.path.abspath('.'))


        # -- Project information -----------------------------------------------------
        import sphinx_rtd_theme
        from os.path import exists
        from os.path import dirname
        from os.path import join


        def parse_version(fpath):
            """
            Statically parse the version number from a python file
            """
            import ast
            if not exists(fpath):
                raise ValueError('fpath={{!r}} does not exist'.format(fpath))
            with open(fpath, 'r') as file_:
                sourcecode = file_.read()
            pt = ast.parse(sourcecode)
            class VersionVisitor(ast.NodeVisitor):
                def visit_Assign(self, node):
                    for target in node.targets:
                        if getattr(target, 'id', None) == '__version__':
                            self.version = node.value.s
            visitor = VersionVisitor()
            visitor.visit(pt)
            return visitor.version

        project = '{repo_name}'
        copyright = '{year}, {author}'
        author = '{author}'
        modname = '{mod_name}'

        modpath = join(dirname(dirname(dirname(__file__))), '{rel_mod_dpath}', '__init__.py')
        release = parse_version(modpath)
        version = '.'.join(release.split('.')[0:2])


        # -- General configuration ---------------------------------------------------

        # If your documentation needs a minimal Sphinx version, state it here.
        #
        # needs_sphinx = '1.0'

        # Add any Sphinx extension module names here, as strings. They can be
        # extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
        # ones.
        extensions = [
            # 'autoapi.extension',
            'sphinx.ext.autodoc',
            'sphinx.ext.autosummary',
            'sphinx.ext.intersphinx',
            'sphinx.ext.napoleon',
            'sphinx.ext.todo',
            'sphinx.ext.viewcode',
            # 'myst_parser',  # TODO

            'sphinx.ext.githubpages',
            # 'sphinxcontrib.redirects',
            'sphinx_reredirects',
        ]

        todo_include_todos = True
        napoleon_google_docstring = True
        napoleon_use_param = False
        napoleon_use_ivar = True

        autodoc_inherit_docstrings = False

        autodoc_member_order = 'bysource'
        autoclass_content = 'both'
        # autodoc_mock_imports = ['torch', 'torchvision', 'visdom']

        # autoapi_modules = {{
        #     modname: {{
        #         'override': False,
        #         'output': 'auto'
        #     }}
        # }}
        # autoapi_dirs = [f'../../src/{{modname}}']
        # autoapi_keep_files = True

        intersphinx_mapping = {{
            # 'pytorch': ('http://pytorch.org/docs/master/', None),
            'python': ('https://docs.python.org/3', None),
            'click': ('https://click.palletsprojects.com/', None),
            # 'xxhash': ('https://pypi.org/project/xxhash/', None),
            # 'pygments': ('https://pygments.org/docs/', None),
            # 'tqdm': ('https://tqdm.github.io/', None),
            # Requries that the repo have objects.inv
            'kwarray': ('https://kwarray.readthedocs.io/en/latest/', None),
            'kwimage': ('https://kwimage.readthedocs.io/en/latest/', None),
            # 'kwplot': ('https://kwplot.readthedocs.io/en/latest/', None),
            'ndsampler': ('https://ndsampler.readthedocs.io/en/latest/', None),
            'ubelt': ('https://ubelt.readthedocs.io/en/latest/', None),
            'xdoctest': ('https://xdoctest.readthedocs.io/en/latest/', None),
            'networkx': ('https://networkx.org/documentation/stable/', None),
            'scriptconfig': ('https://scriptconfig.readthedocs.io/en/latest/', None),
            'rich': ('https://rich.readthedocs.io/en/latest/', None),

            'pytest': ('https://docs.pytest.org/en/latest/', None),
            # 'pytest._pytest.doctest': ('https://docs.pytest.org/en/latest/_modules/_pytest/doctest.html', None),
            # 'colorama': ('https://pypi.org/project/colorama/', None),
            # 'numpy': ('http://docs.scipy.org/doc/numpy/', None),
            # 'cv2' : ('http://docs.opencv.org/2.4/', None),
            # 'h5py' : ('http://docs.h5py.org/en/latest/', None)
        }}
        __dev_note__ = """
        python -m sphinx.ext.intersphinx https://docs.python.org/3/objects.inv
        python -m sphinx.ext.intersphinx https://kwcoco.readthedocs.io/en/latest/objects.inv
        python -m sphinx.ext.intersphinx https://networkx.org/documentation/stable/objects.inv
        python -m sphinx.ext.intersphinx https://kwarray.readthedocs.io/en/latest/objects.inv
        python -m sphinx.ext.intersphinx https://kwimage.readthedocs.io/en/latest/objects.inv
        python -m sphinx.ext.intersphinx https://ubelt.readthedocs.io/en/latest/objects.inv
        python -m sphinx.ext.intersphinx https://networkx.org/documentation/stable/objects.inv

        sphobjinv suggest -t 90 -u https://readthedocs.org/projects/pytest/reference/objects.inv
        "signal.convolve2d"

        python -m sphinx.ext.intersphinx https://pygments-doc.readthedocs.io/en/latest/objects.inv
        """


        # Add any paths that contain templates here, relative to this directory.
        templates_path = ['_templates']

        # The suffix(es) of source filenames.
        # You can specify multiple suffix as a list of string:
        #
        source_suffix = ['.rst', '.md']

        # The master toctree document.
        master_doc = 'index'

        # The language for content autogenerated by Sphinx. Refer to documentation
        # for a list of supported languages.
        #
        # This is also used if you do content translation via gettext catalogs.
        # Usually you set "language" from the command line for these cases.
        language = 'en'

        # List of patterns, relative to source directory, that match files and
        # directories to ignore when looking for source files.
        # This pattern also affects html_static_path and html_extra_path .
        exclude_patterns = []

        # The name of the Pygments (syntax highlighting) style to use.
        pygments_style = 'sphinx'


        # -- Options for HTML output -------------------------------------------------

        # The theme to use for HTML and HTML Help pages.  See the documentation for
        # a list of builtin themes.
        #
        html_theme = 'sphinx_rtd_theme'
        html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]

        # Theme options are theme-specific and customize the look and feel of a theme
        # further.  For a list of options available for each theme, see the
        # documentation.
        #
        html_theme_options = {{
            'collapse_navigation': False,
            'display_version': True,
            # 'logo_only': True,
        }}
        # html_logo = '.static/{repo_name}.svg'
        # html_favicon = '.static/{repo_name}.ico'

        # Add any paths that contain custom static files (such as style sheets) here,
        # relative to this directory. They are copied after the builtin static files,
        # so a file named "default.css" will overwrite the builtin "default.css".
        html_static_path = ['_static']

        # Custom sidebar templates, must be a dictionary that maps document names
        # to template names.
        #
        # The default sidebars (for documents that don't match any pattern) are
        # defined by theme itself.  Builtin themes are using these templates by
        # default: ``['localtoc.html', 'relations.html', 'sourcelink.html',
        # 'searchbox.html']``.
        #
        # html_sidebars = {{}}


        # -- Options for HTMLHelp output ---------------------------------------------

        # Output file base name for HTML help builder.
        htmlhelp_basename = project + 'doc'


        # -- Options for LaTeX output ------------------------------------------------

        latex_elements = {{
            # The paper size ('letterpaper' or 'a4paper').
            #
            # 'papersize': 'letterpaper',

            # The font size ('10pt', '11pt' or '12pt').
            #
            # 'pointsize': '10pt',

            # Additional stuff for the LaTeX preamble.
            #
            # 'preamble': '',

            # Latex figure (float) alignment
            #
            # 'figure_align': 'htbp',
        }}

        # Grouping the document tree into LaTeX files. List of tuples
        # (source start file, target name, title,
        #  author, documentclass [howto, manual, or own class]).
        latex_documents = [
            (master_doc, '{repo_name}.tex', '{repo_name} Documentation',
             '{author}', 'manual'),
        ]


        # -- Options for manual page output ------------------------------------------

        # One entry per manual page. List of tuples
        # (source start file, name, description, authors, manual section).
        man_pages = [
            (master_doc, '{repo_name}', '{repo_name} Documentation',
             [author], 1)
        ]


        # -- Options for Texinfo output ----------------------------------------------

        # Grouping the document tree into Texinfo files. List of tuples
        # (source start file, target name, title, author,
        #  dir menu entry, description, category)
        texinfo_documents = [
            (master_doc, '{repo_name}', '{repo_name} Documentation',
             author, '{repo_name}', 'One line description of project.',
             'Miscellaneous'),
        ]


        # -- Extension configuration -------------------------------------------------
        ''').format(**fmtkw)

    from xcookie import rc

    util_text = rc.resource_fpath('conf_ext.py').read_text()
    if 1:
        util_text = util_text.replace('RENDER_IMAGES = 0', 'RENDER_IMAGES = 1')

    if self.config['repo_name'] == 'kwcoco':
        util_text = util_text.replace('HACK_FOR_KWCOCO = 0', 'HACK_FOR_KWCOCO = 1')

    text = text + '\n' + util_text
    return text



import os
import ubelt as ub
from packaging.version import parse as LooseVersion


class NotFound(Exception):
    pass


class Ambiguous(Exception):
    pass


def _return_one(found):
    if len(found) == 0:
        raise NotFound()
    elif len(found) == 1:
        return found[0]
    else:
        raise Ambiguous


class GitlabRemote:
    """
    pip install python-gitlab

    Ignore:
        load_secrets
        HOST=https://gitlab.kitware.com
        export PRIVATE_GITLAB_TOKEN=$(git_token_for "$HOST")

        ipy

        from xcookie.vcs_remotes import *  # NOQA
        self = GitlabRemote('kwutil', 'computer-vision', 'https://gitlab.kitware.com')


    """
    def __init__(self, proj_name, proj_group, url, visibility='public',
                 private_token='env:PRIVATE_GITLAB_TOKEN'):
        import gitlab  # type: ignore
        self.url = url
        self.proj_name = proj_name
        self.proj_path = proj_name
        self.proj_group = proj_group
        self.visibility = visibility
        if private_token.startswith('env:'):
            private_token = os.environ[private_token[4:]]
        self.gitlab = gitlab.Gitlab(url=self.url, private_token=private_token)

    def auth(self):
        self.gitlab.auth()
        return self

    @property
    def group(self):
        gl = self.gitlab
        # Is there a better way to query?
        groups = gl.groups.list(iterator=True)
        found = [g for g in groups if g.name.lower() == self.proj_group.lower()]
        # if not found:
        #     # allow case insensitivity
        #     found = [g for g in groups if g.name.lower() == self.proj_group.lower()]
        return _return_one(found)

    @property
    def project(self):
        group = self.group
        found = [p for p in group.projects.list(iterator=True)
                 if p.path.lower() == self.proj_path.lower()]
        group_project = _return_one(found)
        project = self.gitlab.projects.get(group_project.id)
        return project

    def new_project(self):
        """
        Ignore:
            from xcookie.vcs_remotes import *  # NOQA
            url = 'https://gitlab.kitware.com'
            proj_name = 'kwimage_ext'
            proj_group = 'computer-vision'
            visibility = 'public'
            self = GitlabRemote(proj_name, proj_group, url).auth()
        """
        # https://docs.gitlab.com/ee/api/projects.html#create-project
        group = self.group
        try:
            self.project
        except NotFound:
            pass
        else:
            raise Exception('project already exist')

        new_proj_data = {
            'name': self.proj_name,
            'path': self.proj_path,
            'namespace_id': group.id,
            'initialize_with_readme': False,
            'visibility': self.visibility,
        }
        new_proj = self.gitlab.projects.create(new_proj_data)

        print(new_proj)

        # Setup protected branches
        self.set_protected_branches()

    def set_protected_branches(self):
        project = self.project

        existing_protected_branches = project.protectedbranches.list()
        expected_protected_branches = [
            'release', 'main',
            'master'
        ]

        existing = {b.name for b in existing_protected_branches}
        missing = [bname for bname in expected_protected_branches
                   if bname not in existing]

        for name in missing:
            # https://docs.gitlab.com/ee/api/protected_branches.html#protect-repository-branches
            project.protectedbranches.create({
                'name': name,
                'allow_force_push': False,
            })
        # if hasattr(gitlab.const, 'AccessLevel'):
        #     maintainer = gitlab.const.AccessLevel.MAINTAINER
        # else:
        #     maintainer = gitlab.const.MAINTAINER_ACCESS
        # perm = [{'access_level': maintainer}]

        if not missing:
            protected_branches = existing_protected_branches
        else:
            protected_branches = project.protectedbranches.list()

        # TODO: figure out how to change access levels
        for branch in protected_branches:
            print('---')
            print('branch = {}'.format(ub.repr2(branch, nl=1)))
            print('branch.push_access_levels = {}'.format(ub.repr2(branch.push_access_levels, nl=1)))
            print('branch.merge_access_levels = {}'.format(ub.repr2(branch.merge_access_levels, nl=1)))
            print('branch.allow_force_push = {}'.format(ub.repr2(branch.allow_force_push, nl=1)))
        # if 0:
        #         # API doesnt directly support this, hack it in
        #         post_data = {
        #             'name': branch.get_id(),
        #             'allowed_to_push': perm,
        #             'allowed_to_merge': perm,
        #         }
        #         # branch.manager.gitlab.http_post(
        #         #     f'/projects/{project.id}/protected_branches',
        #         #     post_data=post_data,
        #         # )


class GithubRemote:
    def __init__(self, proj_name):
        self.proj_name = proj_name

    def new_project(self):
        ub.cmd(f'gh repo create {self.proj_name} --public', verbose=3, system=True)

    def publish_release(self):
        """
        POC for making a release script

        References:
            https://cli.github.com/manual/gh_release_create
        """

        fpath = "CHANGELOG.md"
        version_changelines = _parse_changelog(fpath)

        latest_version, latest_notes = ub.peek(version_changelines.items())
        VERSION = ub.cmd('python -c "import setup; print(setup.VERSION)"')['out'].strip()
        DEPLOY_REMOTE = 'origin'
        TAG_NAME = f'v{VERSION}'
        assert latest_version == LooseVersion(VERSION)

        tag_exists = ub.cmd(f'git rev-parse {TAG_NAME}')['ret'] == 0
        if not tag_exists:
            ub.cmd(f'git tag "{TAG_NAME}" -m "tarball tag {VERSION}"', verbose=2)
            ub.cmd(f'git push --tags {DEPLOY_REMOTE}', verbose=2)

        import tempfile
        release_notes_fpath = ub.Path(tempfile.mktemp('.txt'))
        release_notes_fpath.write_text('\n'.join(latest_notes[1:]))

        title = f'Version {latest_version}'
        command = f'gh release create "{TAG_NAME}" --notes-file "{release_notes_fpath}" --title "{title}"'
        print(command)
        ub.cmd(command, verbose=3)


def version_bump():
    #### Do a version bump on the repo
    # Update the changelog
    VERSION = LooseVersion(ub.cmd('python -c "import setup; print(setup.VERSION)"')['out'].strip())
    DEPLOY_REMOTE = 'origin'
    fpath = "CHANGELOG.md"
    changelog_fpath = ub.Path(fpath)
    NEXT_VERSION = '{}.{}.{}'.format(VERSION.major, VERSION.minor,
                                     VERSION.micro + 1)
    text = changelog_fpath.read_text()
    text = text.replace('Unreleased', 'Released ' + ub.timeparse(ub.timestamp()).date().isoformat())
    lines = text.split(chr(10))
    for ix, line in enumerate(lines):
        if 'Version ' in line:
            break
    newline = fr'## Version {NEXT_VERSION} - Unreleased'
    newlines = lines[:ix] + [newline, '', ''] + lines[ix:]
    new_text = chr(10).join(newlines)
    changelog_fpath.write_text(new_text)

    init_fpath = ub.Path('src/xdoctest/__init__.py')  # hack, hard code
    init_text = init_fpath.read_text()
    init_text = init_text.replace(f'__version__ = {VERSION!r}', f'__version__ = {NEXT_VERSION!r}', )
    init_fpath.write_text(init_text)

    ub.cmd(f'git co -b "dev/{NEXT_VERSION}"')
    ub.cmd(f'git commit -am "Start branch for {NEXT_VERSION}"')
    ub.cmd(f'git push "{DEPLOY_REMOTE}"')

    ub.cmd(f'gh pr create --title "Start branch for {NEXT_VERSION}" --body "auto created PR" --base main --assignee @me', verbose=2)
    # Github create PR


def _parse_changelog(fpath):
    """
    Helper to parse the changelog for the version to verify versions agree.

    CommandLine:
        xdoctest -m dev/parse_changelog.py _parse_changelog --dev
        fpath = "CHANGELOG.md"
    """
    import re
    pat = re.compile(r'#.*Version ([0-9]+\.[0-9]+\.[0-9]+)')
    # We can statically modify this to a constant value when we deploy

    version = None
    versions = {}
    version_changelines = ub.ddict(list)
    with open(fpath, 'r') as file:
        for line in file.readlines():
            line = line.rstrip()
            if line:
                parsed = pat.search(line)
                if parsed:
                    print('parsed = {!r}'.format(parsed))
                    try:
                        version_text = parsed.groups()[0]
                        version = LooseVersion(version_text)
                        versions.append(version)
                    except Exception:
                        print('Failed to parse = {!r}'.format(line))
            if version is not None:
                version_changelines[version].append(line)
    return version_changelines



"""
Common subroutines for consitency between gitlab-ci / github actions / etc...
"""


def make_build_wheel_parts(self, wheelhouse_dpath='wheelhouse'):
    commands = [
        # 'python -m pip install pip -U',
        'python -m pip install setuptools>=0.8 wheel build twine',
        f'python -m build --wheel --outdir {wheelhouse_dpath}',
        f'python -m twine check ./{wheelhouse_dpath}/{self.mod_name}*.whl',
    ]

    build_wheel_parts = {
        'commands': commands,
        'artifact': f"./{wheelhouse_dpath}/{self.mod_name}*.whl"
    }
    return build_wheel_parts


def make_install_and_test_wheel_parts(self,
                                      wheelhouse_dpath,
                                      special_install_lines,
                                      workspace_dname,
                                      test_command='auto',
                                      custom_before_test_lines=[],
                                      custom_after_test_commands=[],
                                      ):
    """
    Builds the YAML common between github actions and gitlab CI to install and
    tests python packages.

    References:
        https://stackoverflow.com/questions/42019184/python-how-can-i-get-the-version-number-from-a-whl-file
    """
    from xcookie.util_yaml import Yaml

    # get_modname_python = "import tomli; print(tomli.load(open('pyproject.toml', 'rb'))['tool']['xcookie']['mod_name'])"
    # get_modname_bash = f'python -c "{get_modname_python}"'

    get_wheel_fpath_python = Yaml.CodeBlock(f"import pathlib; print(str(sorted(pathlib.Path('{wheelhouse_dpath}').glob('{self.mod_name}*.whl'))[-1]).replace(chr(92), chr(47)))")
    # get_wheel_fpath_python = f"import pathlib; print(str(sorted(pathlib.Path('{wheelhouse_dpath}').glob('{self.mod_name}*.whl'))[-1]).replace(r'\\', '/'))"
    get_wheel_fpath_bash = f'python -c "{get_wheel_fpath_python}"'

    get_mod_version_python = "from pkginfo import Wheel; print(Wheel('$WHEEL_FPATH').version)"
    get_mod_version_bash = f'python -c "{get_mod_version_python}"'

    # get_modpath_python = "import ubelt; print(ubelt.modname_to_modpath(f'{self.mod_name}'))"
    get_modpath_python = f"import {self.mod_name}, os; print(os.path.dirname({self.mod_name}.__file__))"
    get_modpath_bash = f'python -c "{get_modpath_python}"'

    if test_command == 'auto':
        test_command = [
            Yaml.CodeBlock(f'python -m pytest --verbose -p pytester -p no:doctest --xdoctest --cov-config ../pyproject.toml --cov-report term --cov="{self.mod_name}" "$MOD_DPATH" ../tests'),
            'echo "pytest command finished, moving the coverage file to the repo root"',
        ]

    # Note: export does not expose the enviornment variable to subsequent jobs.
    install_wheel_commands = [
        'echo "Finding the path to the wheel"',
        f'ls {wheelhouse_dpath} || echo "{wheelhouse_dpath} does not exist"',
        'echo "Installing helpers"',
        # 'pip install pip setuptools>=0.8 setuptools_scm wheel build -U',  # is this necessary?
        'pip install setuptools>=0.8 setuptools_scm wheel build -U',  # is this necessary?
        'pip install tomli pkginfo',
        # 'pip install delorean',
        f'export WHEEL_FPATH=$({get_wheel_fpath_bash})',
        # 'echo "WHEEL_FPATH=$WHEEL_FPATH"',
        f'export MOD_VERSION=$({get_mod_version_bash})',
        # 'echo "MOD_VERSION=$MOD_VERSION"',
    ] + special_install_lines + [
        f'pip install --prefer-binary "{self.mod_name}[$INSTALL_EXTRAS]==$MOD_VERSION" -f {wheelhouse_dpath}',
        'echo "Install finished."',
    ]

    test_wheel_commands = [
        'echo "Creating test sandbox directory"',
        f'export WORKSPACE_DNAME="{workspace_dname}"',
        'echo "WORKSPACE_DNAME=$WORKSPACE_DNAME"',
        'mkdir -p $WORKSPACE_DNAME',
        'echo "cd-ing into the workspace"',
        'cd $WORKSPACE_DNAME',
        'pwd',
        'ls -altr',
        # 'pip freeze',
        '# Get the path to the installed package and run the tests',
        f'export MOD_DPATH=$({get_modpath_bash})',
        Yaml.CodeBlock(
            '''
            echo "
            ---
            MOD_DPATH = $MOD_DPATH
            ---
            running the pytest command inside the workspace
            ---
            "
            '''),
    ] + custom_before_test_lines + test_command + custom_after_test_commands

    install_and_test_wheel_parts = {
        'install_wheel_commands': install_wheel_commands,
        'test_wheel_commands': test_wheel_commands,
    }
    return install_and_test_wheel_parts



from sphinx.domains.python import PythonDomain  # NOQA
# from sphinx.application import Sphinx  # NOQA
from typing import Any, List  # NOQA


class PatchedPythonDomain(PythonDomain):
    """
    References:
        https://github.com/sphinx-doc/sphinx/issues/3866
    """
    def resolve_xref(self, env, fromdocname, builder, typ, target, node, contnode):
        # TODO: can use this to resolve references nicely
        # if target.startswith('ub.'):
        #     target = 'ubelt.' + target[3]
        return_value = super(PatchedPythonDomain, self).resolve_xref(
            env, fromdocname, builder, typ, target, node, contnode)
        return return_value


class GoogleStyleDocstringProcessor:
    """
    A small extension that runs after napoleon and reformats erotemic-flavored
    google-style docstrings for sphinx.
    """

    def __init__(self, autobuild=1):
        self.registry = {}
        if autobuild:
            self._register_builtins()

    def register_section(self, tag, alias=None):
        """
        Decorator that adds a custom processing function for a non-standard
        google style tag. The decorated function should accept a list of
        docstring lines, where the first one will be the google-style tag that
        likely needs to be replaced, and then return the appropriate sphinx
        format (TODO what is the name? Is it just RST?).
        """
        alias = [] if alias is None else alias
        alias = [alias] if not isinstance(alias, (list, tuple, set)) else alias
        alias.append(tag)
        alias = tuple(alias)
        # TODO: better tag patterns
        def _wrap(func):
            self.registry[tag] = {
                'tag': tag,
                'alias': alias,
                'func': func,
            }
            return func
        return _wrap

    def _register_builtins(self):
        """
        Adds definitions I like of CommandLine, TextArt, and Ignore
        """

        @self.register_section(tag='CommandLine')
        def commandline(lines):
            new_lines = []
            new_lines.append('.. rubric:: CommandLine')
            new_lines.append('')
            new_lines.append('.. code-block:: bash')
            new_lines.append('')
            new_lines.extend(lines[1:])
            return new_lines

        @self.register_section(tag='SpecialExample', alias=['Benchmark', 'Sympy', 'Doctest'])
        def benchmark(lines):
            import textwrap
            new_lines = []
            tag = lines[0].replace(':', '').strip()
            # new_lines.append(lines[0])  # TODO: it would be nice to change the tagline.
            # new_lines.append('')
            new_lines.append('.. rubric:: {}'.format(tag))
            new_lines.append('')
            new_text = textwrap.dedent('\n'.join(lines[1:]))
            redone = new_text.split('\n')
            new_lines.extend(redone)
            # import ubelt as ub
            # print('new_lines = {}'.format(ub.repr2(new_lines, nl=1)))
            # new_lines.append('')
            return new_lines

        @self.register_section(tag='TextArt', alias=['Ascii'])
        def text_art(lines):
            new_lines = []
            new_lines.append('.. rubric:: TextArt')
            new_lines.append('')
            new_lines.append('.. code-block:: bash')
            new_lines.append('')
            new_lines.extend(lines[1:])
            return new_lines

        @self.register_section(tag='Ignore')
        def ignore(lines):
            return []

    def process(self, lines):
        """
        Example:
            >>> import ubelt as ub
            >>> self = GoogleStyleDocstringProcessor()
            >>> lines = ['Hello world',
            >>>              '',
            >>>              'CommandLine:',
            >>>              '    hi',
            >>>              '',
            >>>              'CommandLine:',
            >>>              '',
            >>>              '    bye',
            >>>              '',
            >>>              'TextArt:',
            >>>              '',
            >>>              '    1',
            >>>              '    2',
            >>>              '',
            >>>              '    345',
            >>>              '',
            >>>              'Foobar:',
            >>>              '',
            >>>              'TextArt:']
            >>> new_lines = self.process(lines[:])
            >>> print(chr(10).join(new_lines))
        """
        orig_lines = lines[:]
        new_lines = []
        curr_mode = '__doc__'
        accum = []

        def accept():
            """ called when we finish reading a section """
            if curr_mode == '__doc__':
                # Keep the lines as-is
                new_lines.extend(accum)
            else:
                # Process this section with the given function
                regitem = self.registry[curr_mode]
                func = regitem['func']
                fixed = func(accum)
                new_lines.extend(fixed)
            # Reset the accumulator for the next section
            accum[:] = []

        for line in orig_lines:

            found = None
            for regitem in self.registry.values():
                if line.startswith(regitem['alias']):
                    found = regitem['tag']
                    break
            if not found and line and not line.startswith(' '):
                # if the line startswith anything but a space, we are no longer
                # in the previous nested scope. NOTE: This assumption may not
                # be general, but it works for my code.
                found = '__doc__'

            if found:
                # New section is found, accept the previous one and start
                # accumulating the new one.
                accept()
                curr_mode = found

            accum.append(line)

        # Finialize the last section
        accept()

        lines[:] = new_lines
        # make sure there is a blank line at the end
        if lines and lines[-1]:
            lines.append('')

        return lines

    def process_docstring_callback(self, app, what_: str, name: str, obj: Any,
                                   options: Any, lines: List[str]) -> None:
        """
        Callback to be registered to autodoc-process-docstring

        Custom process to transform docstring lines Remove "Ignore" blocks

        Args:
            app (sphinx.application.Sphinx): the Sphinx application object

            what (str):
                the type of the object which the docstring belongs to (one of
                "module", "class", "exception", "function", "method", "attribute")

            name (str): the fully qualified name of the object

            obj: the object itself

            options: the options given to the directive: an object with
                attributes inherited_members, undoc_members, show_inheritance
                and noindex that are true if the flag option of same name was
                given to the auto directive

            lines (List[str]): the lines of the docstring, see above

        References:
            https://www.sphinx-doc.org/en/1.5.1/_modules/sphinx/ext/autodoc.html
            https://www.sphinx-doc.org/en/master/usage/extensions/autodoc.html
        """
        # print(f'name={name}')
        # print('BEFORE:')
        # import ubelt as ub
        # print('lines = {}'.format(ub.repr2(lines, nl=1)))

        self.process(lines)

        # docstr = '\n'.join(lines)
        # if 'Convert the Mask' in docstr:
        #     import xdev
        #     xdev.embed()

        # if 'keys in this dictionary ' in docstr:
        #     import xdev
        #     xdev.embed()

        RENDER_IMAGES = 0
        if RENDER_IMAGES:
            # DEVELOPING
            if any('REQUIRES(--show)' in line for line in lines):
                # import xdev
                # xdev.embed()
                create_doctest_figure(app, obj, name, lines)

        FIX_EXAMPLE_FORMATTING = 1
        if FIX_EXAMPLE_FORMATTING:
            for idx, line in enumerate(lines):
                if line == "Example:":
                    lines[idx] = "**Example:**"
                    lines.insert(idx + 1, "")

        REFORMAT_SECTIONS = 0
        if REFORMAT_SECTIONS:
            REFORMAT_RETURNS = 0
            REFORMAT_PARAMS = 0

            docstr = SphinxDocstring(lines)

            if REFORMAT_PARAMS:
                for found in docstr.find_tagged_lines('Parameters'):
                    print(found['text'])
                    edit_slice = found['edit_slice']

                    # TODO: figure out how to do this.

                    # # file = 'foo.rst'
                    # import rstparse
                    # rst = rstparse.Parser()
                    # import io
                    # rst.read(io.StringIO(found['text']))
                    # rst.parse()
                    # for line in rst.lines:
                    #     print(line)

                    # # found['text']
                    # import docutils

                    # settings = docutils.frontend.OptionParser(
                    #     components=(docutils.parsers.rst.Parser,)
                    #     ).get_default_values()
                    # document = docutils.utils.new_document('<tempdoc>', settings)
                    # from docutils.parsers import rst
                    # rst.Parser().parse(found['text'], document)

            if REFORMAT_RETURNS:
                for found in docstr.find_tagged_lines('returns'):
                    # FIXME: account for new slice with -2 offset
                    edit_slice = found['edit_slice']
                    text = found['text']
                    new_lines = []
                    for para in text.split('\n\n'):
                        indent = para[:len(para) - len(para.lstrip())]
                        new_paragraph = indent + paragraph(para)
                        new_lines.append(new_paragraph)
                        new_lines.append('')
                    new_lines = new_lines[:-1]
                    lines[edit_slice] = new_lines

        # print('AFTER:')
        # print('lines = {}'.format(ub.repr2(lines, nl=1)))

        # if name == 'kwimage.Affine.translate':
        #     import sys
        #     sys.exit(1)


class SphinxDocstring:
    """
    Helper to parse and modify sphinx docstrings
    """
    def __init__(docstr, lines):
        docstr.lines = lines

        # FORMAT THE RETURNS SECTION A BIT NICER
        import re
        tag_pat = re.compile(r'^:(\w*):')
        directive_pat = re.compile(r'^.. (\w*)::\s*(\w*)')

        # Split by sphinx types, mark the line offset where they start / stop
        sphinx_parts = []
        for idx, line in enumerate(lines):
            tag_match = tag_pat.search(line)
            directive_match = directive_pat.search(line)
            if tag_match:
                tag = tag_match.groups()[0]
                sphinx_parts.append({
                    'tag': tag, 'start_offset': idx,
                    'type': 'tag',
                })
            elif directive_match:
                tag = directive_match.groups()[0]
                sphinx_parts.append({
                    'tag': tag, 'start_offset': idx,
                    'type': 'directive',
                })

        prev_offset = len(lines)
        for part in sphinx_parts[::-1]:
            part['end_offset'] = prev_offset
            prev_offset = part['start_offset']

        docstr.sphinx_parts = sphinx_parts

        if 0:
            for line in lines:
                print(line)

    def find_tagged_lines(docstr, tag):
        for part in docstr.sphinx_parts[::-1]:
            if part['tag'] == tag:
                edit_slice = slice(part['start_offset'], part['end_offset'])
                return_section = docstr.lines[edit_slice]
                text = '\n'.join(return_section)
                found = {
                    'edit_slice': edit_slice,
                    'text': text,
                }
                yield found


def paragraph(text):
    r"""
    Wraps multi-line strings and restructures the text to remove all newlines,
    heading, trailing, and double spaces.

    Useful for writing log messages

    Args:
        text (str): typically a multiline string

    Returns:
        str: the reduced text block
    """
    import re
    out = re.sub(r'\s\s*', ' ', text).strip()
    return out


def create_doctest_figure(app, obj, name, lines):
    """
    The idea is that each doctest that produces a figure should generate that
    and then that figure should be part of the docs.
    """
    import xdoctest
    import sys
    import types
    if isinstance(obj, types.ModuleType):
        module = obj
    else:
        module = sys.modules[obj.__module__]
    # TODO: read settings from pyproject.toml?
    if '--show' not in sys.argv:
        sys.argv.append('--show')
    if '--nointeract' not in sys.argv:
        sys.argv.append('--nointeract')
    modpath = module.__file__

    # print(doctest.format_src())
    import pathlib
    # HACK: write to the srcdir
    doc_outdir = pathlib.Path(app.outdir)
    doc_srcdir = pathlib.Path(app.srcdir)
    doc_static_outdir = doc_outdir / '_static'
    doc_static_srcdir = doc_srcdir / '_static'
    src_fig_dpath = (doc_static_srcdir / 'images')
    src_fig_dpath.mkdir(exist_ok=True, parents=True)
    out_fig_dpath = (doc_static_outdir / 'images')
    out_fig_dpath.mkdir(exist_ok=True, parents=True)

    # fig_dpath = (doc_outdir / 'autofigs' / name).mkdir(exist_ok=True)

    fig_num = 1

    import kwplot
    kwplot.autompl(force='agg')
    plt = kwplot.autoplt()

    docstr = '\n'.join(lines)

    # TODO: The freeform parser does not work correctly here.
    # We need to parse out the sphinx (epdoc)? individual examples
    # so we can get different figures. But we can hack it for now.

    import re
    split_parts = re.split('({}\\s*\n)'.format(re.escape('.. rubric:: Example')), docstr)
    # split_parts = docstr.split('.. rubric:: Example')

    # import xdev
    # xdev.embed()

    def doctest_line_offsets(doctest):
        # Where the doctests starts and ends relative to the file
        start_line_offset = doctest.lineno - 1
        last_part = doctest._parts[-1]
        last_line_offset = start_line_offset + last_part.line_offset + last_part.n_lines - 1
        offsets = {
            'start': start_line_offset,
            'end': last_line_offset,
            'stop': last_line_offset + 1,
        }
        return offsets

    # from xdoctest import utils
    # part_lines = utils.add_line_numbers(docstr.split('\n'), n_digits=3, start=0)
    # print('\n'.join(part_lines))

    to_insert_fpaths = []
    curr_line_offset = 0
    for part in split_parts:
        num_lines = part.count('\n')

        doctests = list(xdoctest.core.parse_docstr_examples(
            part, modpath=modpath, callname=name,
            # style='google'
        ))
        # print(doctests)

        # doctests = list(xdoctest.core.parse_docstr_examples(
        #     docstr, modpath=modpath, callname=name))

        for doctest in doctests:
            if '--show' in part:
                ...
                # print('-- SHOW TEST---')/)
                # kwplot.close_figures()
                try:
                    import pytest  # NOQA
                except ImportError:
                    pass
                try:
                    from xdoctest.exceptions import Skipped
                except ImportError:  # nocover
                    # Define dummy skipped exception if pytest is not available
                    class Skipped(Exception):
                        pass
                try:
                    doctest.mode = 'native'
                    doctest.run(verbose=0, on_error='raise')
                    ...
                except Skipped:
                    print(f'Skip doctest={doctest}')
                except Exception as ex:
                    print(f'ex={ex}')
                    print(f'Error in doctest={doctest}')

                offsets = doctest_line_offsets(doctest)
                doctest_line_end = curr_line_offset + offsets['stop']
                insert_line_index = doctest_line_end

                figures = kwplot.all_figures()
                for fig in figures:
                    fig_num += 1
                    # path_name = path_sanatize(name)
                    path_name = (name).replace('.', '_')
                    fig_fpath = src_fig_dpath / f'fig_{path_name}_{fig_num:03d}.jpeg'
                    fig.savefig(fig_fpath)
                    print(f'Wrote figure: {fig_fpath}')
                    to_insert_fpaths.append({
                        'insert_line_index': insert_line_index,
                        'fpath': fig_fpath,
                    })

                for fig in figures:
                    plt.close(fig)
                # kwplot.close_figures(figures)

        curr_line_offset += (num_lines)

    # if len(doctests) > 1:
    #     doctests
    #     import xdev
    #     xdev.embed()

    INSERT_AT = 'end'
    INSERT_AT = 'inline'

    end_index = len(lines)
    # Reverse order for inserts
    import shutil
    for info in to_insert_fpaths[::-1]:
        src_abs_fpath = info['fpath']

        rel_to_static_fpath = src_abs_fpath.relative_to(doc_static_srcdir)
        # dst_abs_fpath = doc_static_outdir / rel_to_static_fpath
        # dst_abs_fpath.parent.mkdir(parents=True, exist_ok=True)

        rel_to_root_fpath = src_abs_fpath.relative_to(doc_srcdir)

        dst_abs_fpath1 = doc_outdir / rel_to_root_fpath
        dst_abs_fpath1.parent.mkdir(parents=True, exist_ok=True)
        shutil.copy(src_abs_fpath, dst_abs_fpath1)

        dst_abs_fpath2 = doc_outdir / rel_to_static_fpath
        dst_abs_fpath2.parent.mkdir(parents=True, exist_ok=True)
        shutil.copy(src_abs_fpath, dst_abs_fpath2)

        dst_abs_fpath3 = doc_srcdir / rel_to_static_fpath
        dst_abs_fpath3.parent.mkdir(parents=True, exist_ok=True)
        shutil.copy(src_abs_fpath, dst_abs_fpath3)

        if INSERT_AT == 'inline':
            # Try to insert after test
            insert_index = info['insert_line_index']
        elif INSERT_AT == 'end':
            insert_index = end_index
        else:
            raise KeyError(INSERT_AT)
        lines.insert(insert_index, '.. image:: {}'.format(rel_to_root_fpath))
        # lines.insert(insert_index, '.. image:: {}'.format(rel_to_static_fpath))
        lines.insert(insert_index, '')


def setup(app):
    import sphinx
    app : sphinx.application.Sphinx = app
    app.add_domain(PatchedPythonDomain, override=True)
    docstring_processor = GoogleStyleDocstringProcessor()
    # https://stackoverflow.com/questions/26534184/can-sphinx-ignore-certain-tags-in-python-docstrings
    app.connect('autodoc-process-docstring', docstring_processor.process_docstring_callback)

    ### Hack for kwcoco: TODO: figure out a way for the user to configure this.
    HACK_FOR_KWCOCO = 0
    if HACK_FOR_KWCOCO:
        import pathlib
        import shutil
        doc_outdir = pathlib.Path(app.outdir)
        doc_srcdir = pathlib.Path(app.srcdir)
        schema_src = (doc_srcdir / '../../kwcoco/coco_schema.json')
        shutil.copy(schema_src, doc_outdir / 'coco_schema.json')
        shutil.copy(schema_src, doc_srcdir / 'coco_schema.json')
    return app



def build_setup(self):
    from xcookie import rc
    import ubelt as ub
    import re
    # from distutils.version import Version
    # from packaging.version import parse as Version

    pkg_name = self.config['pkg_name']
    min_python = self.config['min_python']
    # min_py_version = str(self.config['min_python'])
    dev_status = self.config['dev_status']

    fpath = rc.resource_fpath('setup.py.in')

    template_text = fpath.read_text()
    parts = re.split('(### xcookie: .*$)', template_text, flags=re.MULTILINE)

    lut = {}
    for p in parts:
        if not p.strip():
            continue
        if p.startswith('### xcookie: +'):
            key = p.split('+')[1]
        elif p.startswith('### xcookie: -'):
            assert key in lut
        else:
            assert key not in lut, f'{key=}'
            lut[key] = p

    parts = []
    parts.append(ub.codeblock(
        '''
        #!/usr/bin/env python
        # Generated by ~/code/xcookie/xcookie/builders/setup.py
        # based on part ~/code/xcookie/xcookie/rc/setup.py.in
        import sys
        import re
        from os.path import exists, dirname, join
        from setuptools import find_packages
        '''))

    if 'binpy' in self.tags:
        parts.append(lut['IF(binpy)'])
    else:
        parts.append(ub.codeblock(
            '''
            from setuptools import setup
            '''))

    parts.append(lut['HELPERS'])
    parts.append(ub.codeblock(
        f'''
        NAME = '{pkg_name}'
        INIT_PATH = '{self.rel_mod_dpath}/__init__.py'
        VERSION = parse_version(INIT_PATH)
        '''))

    version_classifiers = []
    for ver in self.config['supported_python_versions']:
        version_classifiers.append(f'Programming Language :: Python :: {ver}')

    # List of classifiers available at:
    dev_status = dev_status.lower()
    if dev_status == 'planning':
        dev_status = 'Development Status :: 1 - Planning'
    elif dev_status == 'pre-alpha':
        dev_status = 'Development Status :: 2 - Pre-Alpha'
    elif dev_status == 'alpha':
        dev_status = 'Development Status :: 3 - Alpha'
    elif dev_status == 'beta':
        dev_status = 'Development Status :: 4 - Beta'
    elif dev_status in {'stable', 'production'}:
        dev_status = 'Development Status :: 5 - Production/Stable'
    elif dev_status == 'mature':
        dev_status = 'Development Status :: 6 - Mature'
    elif dev_status == 'inactive':
        dev_status = 'Development Status :: 7 - Inactive'

    other_classifiers = [
        # https://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Utilities',
        # This should be interpreted as Apache License v2.0
        'License :: OSI Approved :: Apache Software License',
    ]

    pyproject_settings = self.config._load_xcookie_pyproject_settings()
    if 'classifiers' in pyproject_settings:
        other_classifiers += pyproject_settings['classifiers']

    classifiers = [dev_status] + other_classifiers + version_classifiers

    # if 0:
    #     setupkw['entry_points'] = {
    #         # the console_scripts entry point creates the package CLI
    #         'console_scripts': [
    #             'xcookie = xcookie.__main__:main'
    #         ]
    #     }
    parts.append(ub.identity(
        '''
if __name__ == '__main__':
    setupkw = {}

    setupkw['install_requires'] = parse_requirements('requirements/runtime.txt', versions='loose')
    setupkw['extras_require'] = {
        'all': parse_requirements('requirements.txt', versions='loose'),
        'tests': parse_requirements('requirements/tests.txt', versions='loose'),
        'optional': parse_requirements('requirements/optional.txt', versions='loose'),

        '''))

    cv2_part = ub.identity(
        '''
        'headless': parse_requirements('requirements/headless.txt', versions='loose'),
        'graphics': parse_requirements('requirements/graphics.txt', versions='loose'),
        # Strict versions
        'headless-strict': parse_requirements('requirements/headless.txt', versions='strict'),
        'graphics-strict': parse_requirements('requirements/graphics.txt', versions='strict'),
        ''')
    if 'cv2' in self.tags:
        parts.append(cv2_part)

    postgresql_part = '''
        'postgresql': parse_requirements('requirements/postgresql.txt', versions='loose'),
        'postgresql-strict': parse_requirements('requirements/postgresql.txt', versions='strict'),
    '''

    if 'postgresql' in self.tags:
        parts.append(postgresql_part)

    parts.append(ub.identity(
        '''
        'all-strict': parse_requirements('requirements.txt', versions='strict'),
        'runtime-strict': parse_requirements('requirements/runtime.txt', versions='strict'),
        'tests-strict': parse_requirements('requirements/tests.txt', versions='strict'),
        'optional-strict': parse_requirements('requirements/optional.txt', versions='strict'),
        }
        '''))

    classifier_text = ub.urepr(classifiers)

    # author=static_parse('__author__', INIT_PATH),
    # author_email=static_parse('__author_email__', INIT_PATH),
    # url=static_parse('__url__', INIT_PATH),

    # TODO: Try placing most of this into a setup.cfg instead
    setupkw_parts = {}
    setupkw_parts['name'] = 'NAME'
    setupkw_parts['version'] = 'VERSION'
    if isinstance(self.config["author"], list):
        setupkw_parts['author'] = repr(', '.join(self.config["author"]))
    else:
        setupkw_parts['author'] = f'{self.config["author"]!r}'
    if isinstance(self.config["author_email"], list):
        setupkw_parts['author_email'] = repr(', '.join(self.config["author_email"]))
    else:
        setupkw_parts['author_email'] = f'{self.config["author_email"]!r}'
    setupkw_parts['url'] = f'{self.config["url"]!r}'
    setupkw_parts['description'] = f'{self.config["description"]!r}'
    setupkw_parts['long_description'] = 'parse_description()'
    setupkw_parts['long_description_content_type'] = "'text/x-rst'"
    setupkw_parts['license'] = f'{self.config["license"]!r}'
    setupkw_parts['packages'] = f"find_packages({self.config['rel_mod_parent_dpath']!r})"
    setupkw_parts['python_requires'] = f"'>={min_python}'"
    setupkw_parts['classifiers'] = f'{classifier_text}'

    if self.config['typed']:
        setupkw_parts['package_data'] = {
            self.mod_name: ['py.typed', '*.pyi'],
        }

    if self.config['rel_mod_parent_dpath'] != '.':
        # https://codefellows.github.io/sea-python-401d4/lectures/python_packaging_1.html
        # We use a key of an empty string to indicate that the directory we are
        # pointing to should be considered the root. Then the value is src, telling
        # setuptools to use that directory as the root of our source.
        setupkw_parts['package_dir'] = ub.urepr(
            {'': self.config['rel_mod_parent_dpath']}
        )

    # hack
    if pyproject_settings is None:
        pyproject_settings = {}
    if 'entry_points' in pyproject_settings:
        setupkw_parts['entry_points'] = ub.urepr(pyproject_settings['entry_points'])
    if 'scripts' in pyproject_settings:
        setupkw_parts['scripts'] = ub.urepr(pyproject_settings['scripts'])
    if 'package_data' in pyproject_settings:
        setupkw_parts.setdefault('package_data', {})
        setupkw_parts['package_data'].update(pyproject_settings['package_data'])
    if 'setuptools' in pyproject_settings:
        setupkw_parts.update(pyproject_settings['setuptools'])

    for k, v in setupkw_parts.items():
        parts.append(ub.indent(f"setupkw[{k!r}] = {v}"))
    parts.append(ub.indent('setup(**setupkw)'))

    text = '\n'.join(parts)
    # print(text)

    # Its annoying, but other than that insufferable quote issue, black is very
    # good. I have a patch, but I need to find the best way to integrate it.
    # it is kinda nice to have the '"' = hints at autogenerated text convention
    # though. I suppose that is something that can be said for it.
    try:
        import black
        text = black.format_str(text, mode=black.Mode(string_normalization=True))
    except Exception:
        print(text)
        raise
    return text



from importlib import resources as importlib_resources


def resource_fpath(fname):
    """
    Example:
        >>> from xcookie.rc import *  # NOQA
        >>> fname = 'setup.py.in'
        >>> fpath = resource_fpath(fname)
        >>> assert fpath.name == fname
        >>> assert fpath.exists()
    """
    # https://importlib-resources.readthedocs.io/en/latest/using.html#migrating-from-legacy
    import ubelt as ub
    try:
        rc_dpath = ub.Path(importlib_resources.files('xcookie.rc'))
    except Exception:
        # FIXME: does this work on < 3.9?
        rc_dpath = ub.Path(ub.modname_to_modpath('xcookie.rc')).absolute()
    fpath = rc_dpath / fname
    return fpath



import ubelt as ub
import toml


def build_pyproject(self):
    """
    Returns:
        str: templated code
    """
    # data = toml.loads((self.template_dpath / 'pyproject.toml').read_text())
    # print('data = {}'.format(ub.repr2(data, nl=5)))
    pyproj_config = ub.AutoDict()
    # {'tool': {}}
    if 'binpy' in self.config['tags']:
        pyproj_config['build-system']['requires'] = [
            "setuptools>=41.0.1",
            # setuptools_scm[toml]
            # "wheel",
            "scikit-build>=0.11.1",
            "numpy",
            "ninja>=1.10.2",
            "cmake>=3.21.2",
            "cython>=0.29.24",
        ]

        supported_cp_version = []
        for pyver in self.config['supported_python_versions']:
            supported_cp_version.append('cp' + pyver.replace('.', ''))

        wheel_build_patterns = []
        for cpver in supported_cp_version:
            wheel_build_patterns.append(cpver + '-*')

        pyproj_config['tool']['cibuildwheel'].update({
            'build': " ".join(wheel_build_patterns),
            'build-frontend': "build",
            # 'skip': "pp* cp27-* cp34-* cp35-* cp36-* *-musllinux_*",
            'skip': "pp* *-musllinux_*",
            'build-verbosity': 1,
            'test-requires': ["-r requirements/tests.txt"],
            'test-command': "python {project}/run_tests.py"
        })

        if True:
            cibw = pyproj_config['tool']['cibuildwheel']
            req_commands = {
                'linux': [
                    'yum install epel-release lz4 lz4-devel -y',
                ],
                'windows': [
                    'choco install lz4 -y',
                ],
                'macos': [
                    'brew install lz4',
                ]
            }
            for plat in req_commands.keys():
                cmd = ' && '.join(req_commands[plat])
                cibw[plat]['before-all'] = cmd
    else:
        pyproj_config['build-system']['requires'] = [
            "setuptools>=41.0.1",
            # setuptools_scm[toml]
            # "wheel>=0.37.1",
        ]
        pyproj_config['build-system']['build-backend'] = 'setuptools.build_meta'

    WITH_PYTEST_INI = 1
    if WITH_PYTEST_INI:
        xdoctest_style = self.config['xdoctest_style']
        pytest_ini_opts = pyproj_config['tool']['pytest']['ini_options']
        pytest_ini_opts['addopts'] = f"-p no:doctest --xdoctest --xdoctest-style={xdoctest_style} --ignore-glob=setup.py --ignore-glob=dev --ignore-glob=docs"
        pytest_ini_opts['norecursedirs'] = ".git ignore build __pycache__ dev _skbuild docs"
        pytest_ini_opts['filterwarnings'] = [
            "default",
            "ignore:.*No cfgstr given in Cacher constructor or call.*:Warning",
            "ignore:.*Define the __nice__ method for.*:Warning",
            "ignore:.*private pytest class or function.*:Warning",
        ]

    WITH_COVERAGE = 1
    if WITH_COVERAGE:
        pyproj_config['tool']['coverage'].update(toml.loads(ub.codeblock(
            '''
            [run]
            branch = true

            [report]
            exclude_lines =[
                "pragma: no cover",
                ".*  # pragma: no cover",
                ".*  # nocover",
                "def __repr__",
                "raise AssertionError",
                "raise NotImplementedError",
                "if 0:",
                "if trace is not None",
                "verbose = .*",
                "^ *raise",
                "^ *pass *$",
                "if _debug:",
                "if __name__ == .__main__.:",
                ".*if six.PY2:"
            ]

            omit=[
                "{REPO_NAME}/__main__.py",
                "*/setup.py"
            ]
            ''').format(REPO_NAME=self.repo_name)))

    pyproj_config['tool']['mypy']['ignore_missing_imports'] = True

    WITH_XCOOKIE = 1
    if WITH_XCOOKIE:
        options_to_save = [
            'tags',
            'mod_name',
            'repo_name',
            'pkg_name',
            'rel_mod_parent_dpath',
            'os',
            'min_python',
            'version',
            'url',
            'author',
            'author_email',
            'description',
            'license',
            'dev_status',
            'typed',
            'remote_host',
            'remote_group',
        ]
        config_to_save = ub.dict_subset(self.config, options_to_save)
        pyproj_config['tool']['xcookie'].update(config_to_save)

    text = toml.dumps(pyproj_config)
    return text



#!/usr/bin/env python
# Generated by ~/code/xcookie/xcookie/builders/setup.py
# based on part ~/code/xcookie/xcookie/rc/setup.py.in
import sys
from os.path import exists
from setuptools import find_packages
from setuptools import setup


def parse_version(fpath):
    """
    Statically parse the version number from a python file
    """
    value = static_parse('__version__', fpath)
    return value


def static_parse(varname, fpath):
    """
    Statically parse the a constant variable from a python file
    """
    import ast

    if not exists(fpath):
        raise ValueError('fpath={!r} does not exist'.format(fpath))
    with open(fpath, 'r') as file_:
        sourcecode = file_.read()
    pt = ast.parse(sourcecode)

    class StaticVisitor(ast.NodeVisitor):
        def visit_Assign(self, node):
            for target in node.targets:
                if getattr(target, 'id', None) == varname:
                    self.static_value = node.value.s

    visitor = StaticVisitor()
    visitor.visit(pt)
    try:
        value = visitor.static_value
    except AttributeError:
        import warnings

        value = 'Unknown {}'.format(varname)
        warnings.warn(value)
    return value


def parse_description():
    """
    Parse the description in the README file

    CommandLine:
        pandoc --from=markdown --to=rst --output=README.rst README.md
        python -c "import setup; print(setup.parse_description())"
    """
    from os.path import dirname, join, exists

    readme_fpath = join(dirname(__file__), 'README.rst')
    # This breaks on pip install, so check that it exists.
    if exists(readme_fpath):
        with open(readme_fpath, 'r') as f:
            text = f.read()
        return text
    return ''


def parse_requirements(fname='requirements.txt', versions='loose'):
    """
    Parse the package dependencies listed in a requirements file but strips
    specific versioning information.

    Args:
        fname (str): path to requirements file
        versions (bool | str, default=False):
            If true include version specs.
            If strict, then pin to the minimum version.

    Returns:
        List[str]: list of requirements items
    """
    from os.path import exists, dirname, join
    import re

    require_fpath = fname

    def parse_line(line, dpath=''):
        """
        Parse information from a line in a requirements text file

        line = 'git+https://a.com/somedep@sometag#egg=SomeDep'
        line = '-e git+https://a.com/somedep@sometag#egg=SomeDep'
        """
        # Remove inline comments
        comment_pos = line.find(' #')
        if comment_pos > -1:
            line = line[:comment_pos]

        if line.startswith('-r '):
            # Allow specifying requirements in other files
            target = join(dpath, line.split(' ')[1])
            for info in parse_require_file(target):
                yield info
        else:
            # See: https://www.python.org/dev/peps/pep-0508/
            info = {'line': line}
            if line.startswith('-e '):
                info['package'] = line.split('#egg=')[1]
            else:
                if '--find-links' in line:
                    # setuptools doesnt seem to handle find links
                    line = line.split('--find-links')[0]
                if ';' in line:
                    pkgpart, platpart = line.split(';')
                    # Handle platform specific dependencies
                    # setuptools.readthedocs.io/en/latest/setuptools.html
                    # #declaring-platform-specific-dependencies
                    plat_deps = platpart.strip()
                    info['platform_deps'] = plat_deps
                else:
                    pkgpart = line
                    platpart = None

                # Remove versioning from the package
                pat = '(' + '|'.join(['>=', '==', '>']) + ')'
                parts = re.split(pat, pkgpart, maxsplit=1)
                parts = [p.strip() for p in parts]

                info['package'] = parts[0]
                if len(parts) > 1:
                    op, rest = parts[1:]
                    version = rest  # NOQA
                    info['version'] = (op, version)
            yield info

    def parse_require_file(fpath):
        dpath = dirname(fpath)
        with open(fpath, 'r') as f:
            for line in f.readlines():
                line = line.strip()
                if line and not line.startswith('#'):
                    for info in parse_line(line, dpath=dpath):
                        yield info

    def gen_packages_items():
        if exists(require_fpath):
            for info in parse_require_file(require_fpath):
                parts = [info['package']]
                if versions and 'version' in info:
                    if versions == 'strict':
                        # In strict mode, we pin to the minimum version
                        if info['version']:
                            # Only replace the first >= instance
                            verstr = ''.join(info['version']).replace('>=', '==', 1)
                            parts.append(verstr)
                    else:
                        parts.extend(info['version'])
                if not sys.version.startswith('3.4'):
                    # apparently package_deps are broken in 3.4
                    plat_deps = info.get('platform_deps')
                    if plat_deps is not None:
                        parts.append(';' + plat_deps)
                item = ''.join(parts)
                yield item

    packages = list(gen_packages_items())
    return packages


# # Maybe use in the future? But has private deps
# def parse_requirements_alt(fpath='requirements.txt', versions='loose'):
#     """
#     Args:
#         versions (str): can be
#             False or "free" - remove all constraints
#             True or "loose" - use the greater or equal (>=) in the req file
#             strict - replace all greater equal with equals
#     """
#     # Note: different versions of pip might have different internals.
#     # This may need to be fixed.
#     from pip._internal.req import parse_requirements
#     from pip._internal.network.session import PipSession
#     requirements = []
#     for req in parse_requirements(fpath, session=PipSession()):
#         if not versions or versions == 'free':
#             req_name = req.requirement.split(' ')[0]
#             requirements.append(req_name)
#         elif versions == 'loose' or versions is True:
#             requirements.append(req.requirement)
#         elif versions == 'strict':
#             part1, *rest = req.requirement.split(';')
#             strict_req = ';'.join([part1.replace('>=', '==')] + rest)
#             requirements.append(strict_req)
#         else:
#             raise KeyError(versions)
#     requirements = [r.replace(' ', '') for r in requirements]
#     return requirements


NAME = 'xcookie'
INIT_PATH = 'xcookie/__init__.py'
VERSION = parse_version('xcookie/__init__.py')

if __name__ == '__main__':
    setupkw = {}

    setupkw['install_requires'] = parse_requirements('requirements/runtime.txt')
    setupkw['extras_require'] = {
        'all': parse_requirements('requirements.txt'),
        'tests': parse_requirements('requirements/tests.txt'),
        'optional': parse_requirements('requirements/optional.txt'),
        'all-strict': parse_requirements('requirements.txt', versions='strict'),
        'runtime-strict': parse_requirements(
            'requirements/runtime.txt', versions='strict'
        ),
        'tests-strict': parse_requirements('requirements/tests.txt', versions='strict'),
        'optional-strict': parse_requirements(
            'requirements/optional.txt', versions='strict'
        ),
    }

    setupkw['name'] = NAME
    setupkw['version'] = VERSION
    setupkw['author'] = 'Jon Crall'
    setupkw['author_email'] = 'erotemic@gmail.com'
    setupkw['url'] = 'https://github.com/Erotemic/xcookie'
    setupkw['description'] = 'The xcookie cookie-cutter Module'
    setupkw['long_description'] = parse_description()
    setupkw['long_description_content_type'] = 'text/x-rst'
    setupkw['license'] = 'Apache 2'
    setupkw['packages'] = find_packages('.')
    setupkw['python_requires'] = '>=3.7'
    setupkw['classifiers'] = [
        'Development Status :: 1 - Planning',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Utilities',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
    ]
    setupkw['package_data'] = {'xcookie': ['py.typed', '*.pyi'], 'xcookie.rc': ['*.in']}
    setupkw['entry_points'] = {
        'console_scripts': [
            'xcookie=xcookie.__main__:main',
        ],
    }
    setup(**setupkw)



"""
Build the readthedocs yaml file.
"""


def build_readthedocs(self):
    from xcookie.util_yaml import Yaml
    import ubelt as ub

    default_data = Yaml.loads(ub.codeblock(
        f'''
        # .readthedocs.yml
        # Read the Docs configuration file
        # See https://docs.readthedocs.io/en/stable/config-file/v2.html for details
        #
        # See Also:
        # https://readthedocs.org/dashboard/{self.repo_name}/advanced/

        # Required
        version: 2

        # Build documentation in the docs/ directory with Sphinx
        sphinx:
          configuration: docs/source/conf.py

        # Build documentation with MkDocs
        #mkdocs:
        #  configuration: mkdocs.yml

        # Optionally build your docs in additional formats such as PDF and ePub
        formats: all

        python:
          version: 3.7
          # Optionally set the version of Python and requirements required to
          # build your docs
          install:
            - requirements: requirements/docs.txt
            - method: pip
              path: .
              #extra_requirements:
              #  - docs

        #conda:
        #  environment: environment.yml
        '''))

    data = default_data.copy()

    if 'cv2' in self.tags:
        data['python']['install'].insert(0, {'requirements': 'requirements/headless.txt'})

    if 'gdal' in self.tags:
        data['python']['install'].insert(0, {'requirements': 'requirements/gdal.txt'})

    import ruamel
    text = ruamel.yaml.round_trip_dump(data, Dumper=ruamel.yaml.RoundTripDumper)
    return text
